import os
import re
import sys
import urllib
import webbrowser
from functools import partial
from typing import Dict, Optional

from airunner.gui.windows.wayland_helper import (
    enable_wayland_window_decorations,
)
import requests
from PIL import Image
from PySide6 import QtGui
from PySide6.QtCore import (
    Slot,
    Signal,
    QProcess,
    QTimer,
    QCoreApplication,
)
from PySide6.QtGui import QGuiApplication, QKeySequence, QAction, QCursor
from PySide6.QtWidgets import (
    QApplication,
    QMainWindow,
    QMessageBox,
    QCheckBox,
    QInputDialog,
    QSystemTrayIcon,
    QMenu,
)
from PySide6.QtGui import QIcon

from bs4 import BeautifulSoup

from airunner.settings import (
    AIRUNNER_STATUS_ERROR_COLOR,
    AIRUNNER_STATUS_NORMAL_COLOR_LIGHT,
    AIRUNNER_STATUS_NORMAL_COLOR_DARK,
    AIRUNNER_NSFW_CONTENT_DETECTED_MESSAGE,
    AIRUNNER_DISCORD_URL,
    AIRUNNER_BASE_PATH,
    AIRUNNER_BUG_REPORT_LINK,
    AIRUNNER_VULNERABILITY_REPORT_LINK,
    AIRUNNER_ART_ENABLED,
    AIRUNNER_MEM_SD_DEVICE,
    AIRUNNER_MEM_LLM_DEVICE,
    AIRUNNER_MEM_TTS_DEVICE,
    AIRUNNER_MEM_STT_DEVICE,
)
from airunner.utils.settings import get_qsettings
from airunner.handlers.llm.agent.actions.bash_execute import bash_execute
from airunner.handlers.llm.agent.actions.show_path import show_path
from airunner.handlers.llm.llm_request import LLMRequest
from airunner.data.models.shortcut_keys import ShortcutKeys
from airunner.data.models.application_settings import ApplicationSettings
from airunner.data.models.image_filter import ImageFilter
from airunner.data.models.drawingpad_settings import DrawingPadSettings
from airunner.data.models.tab import Tab
from airunner.app_installer import AppInstaller
from airunner.enums import (
    SignalCode,
    CanvasToolName,
    GeneratorSection,
    LLMActionType,
    ModelType,
    ModelStatus,
)
from airunner.utils.application.mediator_mixin import MediatorMixin
from airunner.utils.application.get_version import get_version
from airunner.utils.application.create_worker import create_worker
from airunner.utils.widgets import (
    save_splitter_settings,
    load_splitter_settings,
)
from airunner.utils.image import convert_image_to_binary
from airunner.gui.styles.styles_mixin import StylesMixin
from airunner.gui.widgets.stats.stats_widget import StatsWidget
from airunner.gui.widgets.status.status_widget import StatusWidget
from airunner.gui.windows.about.about import AboutWindow
from airunner.gui.windows.filter_window import FilterWindow
from airunner.gui.windows.main.ai_model_mixin import AIModelMixin
from airunner.gui.windows.main.pipeline_mixin import PipelineMixin
from airunner.gui.windows.main.settings_mixin import SettingsMixin
from airunner.gui.windows.main.templates.main_window_ui import Ui_MainWindow
from airunner.gui.windows.prompt_browser.prompt_browser import PromptBrowser
from airunner.gui.windows.settings.airunner_settings import SettingsWindow
from airunner.gui.windows.update.update_window import UpdateWindow
from airunner.gui.managers.icon_manager import IconManager
from airunner.plugin_loader import PluginLoader
from airunner.workers.llm_generate_worker import LLMGenerateWorker
from airunner.workers.mask_generator_worker import MaskGeneratorWorker
from airunner.workers.sd_worker import SDWorker

try:
    from airunner.workers.audio_capture_worker import AudioCaptureWorker
    from airunner.workers.audio_processor_worker import AudioProcessorWorker
    from airunner.workers.tts_generator_worker import TTSGeneratorWorker
    from airunner.workers.tts_vocalizer_worker import TTSVocalizerWorker
except OSError as e:
    print("Error loading audio workers:", e)
    AudioCaptureWorker = None
    AudioProcessorWorker = None
    TTSGeneratorWorker = None
    TTSVocalizerWorker = None


class MainWindow(
    MediatorMixin,
    SettingsMixin,
    StylesMixin,
    PipelineMixin,
    AIModelMixin,
    QMainWindow,
):
    show_grid_toggled = Signal(bool)
    image_generated = Signal(bool)
    generator_tab_changed_signal = Signal()
    load_image = Signal(str)
    load_image_object = Signal(object)
    loaded = Signal()
    window_opened = Signal()
    ui_class_ = Ui_MainWindow
    _window_title = f"AI Runner"
    icons = [
        ("settings", "actionSettings"),
        ("crosshair", "actionToggle_Controlnet"),
        ("cpu", "actionToggle_LLM"),
        ("mic", "actionToggle_Speech_to_Text"),
        ("image", "actionToggle_Stable_Diffusion"),
        ("image", "menuArt"),
        ("message-circle", "menuChat"),
        ("refresh-cw", "actionReset_Settings_2"),
        ("x-circle", "actionQuit"),
        ("plus-circle", "artActionNew"),
        ("upload-icon", "actionImport_image"),
        ("download-icon", "actionExport_image_button"),
        ("message-circle", "actionNew_Conversation"),
        ("trash-2", "actionDelete_conversation"),
        ("scissors", "actionCut"),
        ("copy", "actionCopy"),
        ("clipboard", "actionPaste"),
        ("corner-right-down", "actionRotate_90_clockwise"),
        ("corner-left-down", "actionRotate_90_counter_clockwise"),
        ("delete", "actionClear_all_prompts"),
        ("settings", "actionSettings"),
        ("book-open", "actionPrompt_Browser"),
        ("folder", "actionBrowse_AI_Runner_Path"),
        ("folder", "actionBrowse_Images_Path_2"),
        ("speaker", "actionToggle_Text_to_Speech"),
        ("object-selected-icon", "actionToggle_Active_Grid_Area"),
        ("pencil-icon", "actionToggle_Brush"),
        ("eraser-icon", "actionToggle_Eraser"),
        ("grid", "actionToggle_Grid"),
        ("image", "menuStable_Diffusion"),
        ("activity", "actionStats"),
        ("zap", "actionRun_setup_wizard_2"),
        ("slash", "actionSafety_Checker"),
        ("external-link", "actionBug_report"),
        ("external-link", "actionReport_vulnerability"),
        ("message-square", "actionDiscord"),
        ("download", "actionImport_image"),
        ("upload", "actionExport_image_button"),
        ("codesandbox", "menuWorkflow"),
        ("trash-2", "workflow_actionClear"),
        ("edit", "workflow_actionEdit"),
        ("folder", "workflow_actionOpen"),
        ("pause-circle", "workflow_actionPause"),
        ("play", "workflow_actionRun"),
        ("save", "workflow_actionSave"),
        ("stop-circle", "workflow_actionStop"),
        ("corner-up-right", "actionRedo"),
        ("corner-up-left", "actionUndo"),
        ("save", "actionSave_As"),
        ("target", "actionRecenter"),
    ]

    def __init__(self, *args, **kwargs):
        self.ui = self.ui_class_()
        self.qsettings = get_qsettings()
        self.icon_manager: Optional[IconManager] = None
        self.tab_backup = {}
        self.workflow_tab = None
        self.quitting = False
        self.update_popup = None
        self._document_path = None
        self.prompt = None
        self.negative_prompt = None
        self.image_path = None
        self.token_signal = Signal(str)
        self.api = None
        self.input_event_manager = None
        self.tqdm_callback_triggered = False
        self.action = GeneratorSection.TXT2IMG.value
        self.progress_bar_started = False
        self.window = None
        self.canvas = None
        self.models = None
        self.client = None
        self._version = None
        self._latest_version = None
        self.status_error_color = AIRUNNER_STATUS_ERROR_COLOR
        self.status_normal_color_light = AIRUNNER_STATUS_NORMAL_COLOR_LIGHT
        self.status_normal_color_dark = AIRUNNER_STATUS_NORMAL_COLOR_DARK
        self._themes = None
        self.button_clicked_signal = Signal(Dict)
        self.status_widget = None
        self.header_widget_spacer = None
        self.deterministic_window = None
        self.generator = None
        self._generator = None
        self._generator_settings = None
        self.listening = False
        self.initialized = False
        self._model_status = {
            model_type: ModelStatus.UNLOADED for model_type in ModelType
        }
        self.signal_handlers = {
            SignalCode.SD_SAVE_PROMPT_SIGNAL: self.on_save_stablediffusion_prompt_signal,
            SignalCode.QUIT_APPLICATION: self.handle_quit_application_signal,
            SignalCode.SD_NSFW_CONTENT_DETECTED_SIGNAL: self.on_nsfw_content_detected_signal,
            SignalCode.BASH_EXECUTE_SIGNAL: self.on_bash_execute_signal,
            SignalCode.WRITE_FILE: self.on_write_file_signal,
            SignalCode.TOGGLE_FULLSCREEN_SIGNAL: self.on_toggle_fullscreen_signal,
            SignalCode.TOGGLE_TTS_SIGNAL: self.on_toggle_tts,
            SignalCode.TOGGLE_SD_SIGNAL: self.on_toggle_sd,
            SignalCode.TOGGLE_LLM_SIGNAL: self.on_toggle_llm,
            SignalCode.UNLOAD_NON_SD_MODELS: self.on_unload_non_sd_models,
            SignalCode.LOAD_NON_SD_MODELS: self.on_load_non_sd_models,
            SignalCode.APPLICATION_RESET_SETTINGS_SIGNAL: self._action_reset_settings,
            SignalCode.APPLICATION_RESET_PATHS_SIGNAL: self.on_reset_paths_signal,
            SignalCode.MODEL_STATUS_CHANGED_SIGNAL: self.on_model_status_changed_signal,
            SignalCode.KEYBOARD_SHORTCUTS_UPDATED: self.on_keyboard_shortcuts_updated,
            SignalCode.REFRESH_STYLESHEET_SIGNAL: self.on_theme_changed_signal,
            SignalCode.AI_MODELS_SAVE_OR_UPDATE_SIGNAL: self.on_ai_models_save_or_update_signal,
            SignalCode.NAVIGATE_TO_URL: self.on_navigate_to_url,
            SignalCode.MISSING_REQUIRED_MODELS: self.display_missing_models_error,
            SignalCode.ENABLE_WORKFLOWS_TOGGLED: self.on_enable_workflows_toggled,
            SignalCode.RETRANSLATE_UI_SIGNAL: self.on_retranslate_ui_signal,
        }
        self.logger.debug("Starting AI Runnner")
        super().__init__()
        enable_wayland_window_decorations(self)
        ApplicationSettings.objects.update(
            self.application_settings.id,
            sd_enabled=False,
            llm_enabled=False,
            tts_enabled=False,
            stt_enabled=False,
            controlnet_enabled=False,
        )

        self.single_click_timer = QTimer(self)
        self.single_click_timer.setSingleShot(True)
        self.single_click_timer.timeout.connect(self.handle_single_click)

        # Add plugins directory to Python path
        plugins_path = os.path.join(self.path_settings.base_path, "plugins")
        if plugins_path not in sys.path:
            sys.path.append(plugins_path)

        self._updating_settings = True
        self._updating_settings = False
        self._worker_manager = None
        self.initialize_ui()
        self._initialize_workers()
        self.last_tray_click_time = 0
        self.settings_window = None

    @property
    def generator_tab_widget(self):
        return self.ui.generator_widget

    @property
    def version(self):
        if self._version is None:
            self._version = get_version()
        return f"v{self._version}"

    @property
    def latest_version(self):
        return self._latest_version

    @latest_version.setter
    def latest_version(self, val):
        self._latest_version = val

    @property
    def document_name(self):
        return "Untitled"

    @property
    def enable_workflows(self) -> bool:
        return self.qsettings.value("enable_workflows") == "true"

    """
    Slot functions
    
    The following functions are defined in and connected to the appropriate
    signals in the corresponding ui file.
    """

    def on_tray_icon_activated(self, reason: QSystemTrayIcon.ActivationReason):
        """Handle tray icon activation events."""
        print(f"Tray icon activated with reason: {reason}")

        # For single left click or double click, toggle visibility
        if (
            reason is QSystemTrayIcon.ActivationReason.Trigger
            or reason is QSystemTrayIcon.ActivationReason.DoubleClick
        ):
            print(f"Left click detected (reason: {reason})")
            self.toggle_window_visibility()
        elif reason == QSystemTrayIcon.ActivationReason.MiddleClick:
            print("Middle click detected")
        elif reason == QSystemTrayIcon.ActivationReason.Context:
            print("Context menu requested")
            # Update menu text before showing context menu
            if self.isVisible():
                self.toggle_visibility_action.setText("Hide Window")
            else:
                self.toggle_visibility_action.setText("Show Window")

    @Slot()
    def on_actionQuit_triggered(self):
        self.handle_close()

    @Slot(bool)
    def on_actionToggle_Eraser_toggled(self, active: bool):
        self.toggle_tool(CanvasToolName.ERASER, active)

    @Slot()
    def on_actionRecenter_triggered(self):
        self.api.art.canvas.recenter_grid()

    @Slot()
    def on_actionReset_Settings_2_triggered(self):
        self._action_reset_settings()

    @Slot()
    def import_controlnet_image(self):
        pass

    @Slot()
    def export_controlnet_image(self):
        pass

    @Slot()
    def import_drawingpad_image(self):
        pass

    @Slot()
    def export_drawingpad_image(self):
        pass

    @Slot()
    def on_actionExport_image_button_triggered(self):
        self.api.art.canvas.export_image()

    @Slot()
    def on_actionImport_image_triggered(self):
        self.api.art.canvas.import_image()

    @Slot()
    def on_artActionNew_triggered(self):
        self.api.art.canvas.clear()

    @Slot()
    def on_actionUndo_triggered(self):
        self.api.art.canvas.undo()

    @Slot()
    def on_actionRedo_triggered(self):
        self.api.art.canvas.redo()

    @Slot()
    def on_actionPaste_triggered(self):
        self.api.art.canvas.paste_image()

    @Slot()
    def on_actionCopy_triggered(self):
        self.api.art.canvas.copy_image()

    @Slot()
    def on_actionCut_triggered(self):
        self.api.art.canvas.cut_image()

    @Slot()
    def on_actionRotate_90_clockwise_triggered(self):
        self.api.art.canvas.rotate_image_90_clockwise()

    @Slot()
    def on_actionRotate_90_counter_clockwise_triggered(self):
        self.api.art.canvas.rotate_image_90_counterclockwise()

    @Slot()
    def on_actionClear_all_prompts_triggered(self):
        self.clear_all_prompts()

    @Slot()
    def on_actionBrowse_AI_Runner_Path_triggered(self):
        path = self.path_settings.base_path
        if path == "":
            path = AIRUNNER_BASE_PATH
        show_path(path)

    @Slot()
    def action_show_model_path_txt2img(self):
        self.show_settings_path("txt2img_model_path")

    @Slot()
    def action_show_model_path_inpaint(self):
        self.show_settings_path("inpaint_model_path")

    @Slot()
    def action_show_model_path_embeddings(self):
        self.show_settings_path("embeddings_model_path")

    @Slot()
    def action_show_model_path_lora(self):
        self.show_settings_path("lora_model_path")

    @Slot()
    def action_show_llm(self):
        pass

    @Slot()
    def on_actionReport_vulnerability_triggered(self):
        webbrowser.open(AIRUNNER_VULNERABILITY_REPORT_LINK)

    @Slot()
    def on_actionBug_report_triggered(self):
        webbrowser.open(AIRUNNER_BUG_REPORT_LINK)

    @Slot()
    def on_actionDiscord_triggered(self):
        if AIRUNNER_DISCORD_URL:
            webbrowser.open(AIRUNNER_DISCORD_URL)

    @Slot(bool)
    def action_toggle_mask_layer(self, val: bool):
        if val is True and self.drawing_pad_mask is None:
            self._generate_drawingpad_mask()
        self.update_drawing_pad_settings("mask_layer_enabled", val)
        self.api.art.canvas.mask_layer_toggled()

    @Slot(bool)
    def action_outpaint_toggled(self, val: bool):
        self.update_outpaint_settings("enabled", val)

    @Slot()
    def action_outpaint_export(self):
        pass

    @Slot()
    def action_outpaint_import(self):
        pass

    @Slot()
    def on_actionRun_setup_wizard_2_triggered(self):
        self.show_setup_wizard()

    @Slot(bool)
    def on_actionToggle_Stable_Diffusion_toggled(self, val: bool):
        self.on_toggle_sd({"enabled": val})

    @Slot()
    def on_actionSettings_triggered(self):
        if self.settings_window is None:
            self.settings_window = SettingsWindow(
                prevent_always_on_top=False, exec=False
            )
            self.settings_window.show()
        elif not self.settings_window.isVisible():
            self.settings_window.show()
        self.settings_window.raise_()

    @Slot()
    def on_actionBrowse_Images_Path_2_triggered(self):
        self.show_settings_path("image_path")

    @Slot()
    def on_actionPrompt_Browser_triggered(self):
        PromptBrowser()

    @Slot()
    def on_actionStats_triggered(self):
        widget = StatsWidget()
        # display in a window
        widget.show()

    @Slot(bool)
    def on_actionToggle_Active_Grid_Area_toggled(self, val: bool):
        self.toggle_tool(CanvasToolName.ACTIVE_GRID_AREA, val)

    @Slot(bool)
    def on_actionToggle_Brush_toggled(self, val: bool):
        self.toggle_tool(CanvasToolName.BRUSH, val)

    @Slot(bool)
    def on_actionToggle_Eraser_toggled(self, val: bool):
        self.toggle_tool(CanvasToolName.ERASER, val)

    @Slot(bool)
    def on_actionToggle_Grid_toggled(self, val: bool):
        self.update_grid_settings("show_grid", val)
        self.api.art.canvas.toggle_grid(val)

    @Slot(bool)
    def on_actionToggle_LLM_toggled(self, val: bool):
        self.on_toggle_llm(val=val)

    @Slot(bool)
    def on_actionSafety_Checker_toggled(self, val: bool):
        if val is False:
            self.show_nsfw_warning_popup()
        else:
            self.update_application_settings("nsfw_filter", val)
            self.toggle_nsfw_filter()
            self.api.art.load_safety_checker()

    @Slot(bool)
    def on_actionToggle_Speech_to_Text_toggled(self, val: bool):
        if self._model_status[ModelType.STT] is ModelStatus.LOADING:
            val = not val
        self._update_action_button(
            ModelType.STT,
            self.ui.actionToggle_Speech_to_Text,
            val,
            SignalCode.STT_LOAD_SIGNAL,
            SignalCode.STT_UNLOAD_SIGNAL,
            "stt_enabled",
        )
        QApplication.processEvents()
        self.update_application_settings("stt_enabled", val)

    @Slot()
    def on_workflow_actionClear_triggered(self):
        self.ui.graph.clear_graph()

    @Slot()
    def on_workflow_actionRun_triggered(self):
        self.ui.graph.run_workflow()

    @Slot()
    def on_workflow_actionEdit_triggered(self):
        self.ui.graph.edit_workflow()

    @Slot()
    def on_workflow_actionPause_triggered(self):
        self.ui.graph.pause_workflow()

    @Slot()
    def on_workflow_actionSave_triggered(self):
        self.ui.graph.save_workflow()

    @Slot()
    def on_workflow_actionStop_triggered(self):
        self.ui.graph.stop_workflow()

    @Slot()
    def on_workflow_actionOpen_triggered(self):
        self.ui.graph.load_workflow()

    @Slot(bool)
    def on_actionToggle_Text_to_Speech_toggled(self, val: bool):
        self.on_toggle_tts(val=val)

    @Slot()
    def on_actionAbout_triggered(self):
        AboutWindow()

    @Slot(bool)
    def on_actionToggle_Controlnet_toggled(self, val: bool):
        self.update_controlnet_settings("enabled", val)
        self._update_action_button(
            ModelType.CONTROLNET,
            self.ui.actionToggle_Controlnet,
            val,
            SignalCode.CONTROLNET_LOAD_SIGNAL,
            SignalCode.CONTROLNET_UNLOAD_SIGNAL,
            "controlnet_enabled",
        )

    @Slot()
    def on_actionNew_Conversation_triggered(self):
        self.api.llm.clear_history()

    @Slot()
    def on_actionDelete_conversation_triggered(self):
        current_conversation = self.llm_generator_settings.current_conversation
        self.api.llm.converation_deleted(current_conversation.id)

    def _action_reset_settings(self):
        reply = QMessageBox.question(
            self,
            "Reset Settings",
            "Are you sure you want to reset all settings to their default values?",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )

        if reply == QMessageBox.Yes:
            self.reset_settings()
            self.restart()

    """
    End slot functions
    """

    @staticmethod
    def download_url(url, save_path):
        response = requests.get(url)
        soup = BeautifulSoup(response.content, "html.parser")
        title = soup.title.string if soup.title else url
        # Truncate title to 10 words
        title_words = title.split()[:10]
        filename = "_".join(title_words) + ".html"
        # Replace any characters in filename that are not alphanumerics, underscores, or hyphens
        filename = re.sub(r"[^\w\-_]", "_", filename)
        save_path = os.path.join(save_path, filename)
        with open(save_path, "wb") as file:
            file.write(response.content)
        return filename

    @staticmethod
    def download_pdf(url, save_path):
        response = requests.get(url)
        filename = url.split("/")[-1]  # Get the filename from the URL
        save_path = os.path.join(save_path, filename)
        with open(save_path, "wb") as file:
            file.write(response.content)
        return filename

    def on_navigate_to_url(self, _data: Dict = None):
        url, ok = QInputDialog.getText(self, "Browse Web", "Enter your URL:")
        if ok:
            try:
                result = urllib.parse.urlparse(url)
                is_url = all([result.scheme, result.netloc])
            except ValueError:
                is_url = False

            # If the input is a URL, download it
            if is_url:
                if url.lower().endswith(".pdf"):
                    # Handle PDF file
                    filepath = os.path.expanduser(self.path_settings.pdf_path)
                    filename = self.download_pdf(url, filepath)
                else:
                    # Handle URL
                    filepath = os.path.expanduser(
                        self.path_settings.webpages_path
                    )
                    filename = self.download_url(url, filepath)
            elif os.path.isfile(url):
                filepath = os.path.dirname(url)
                filename = os.path.basename(url)
            else:
                self.logger.error(f"Invalid URL or file path")
                return

            # Update target files to use only the file that was downloaded or navigated to
            # and update the index.
            self.update_chatbot(
                "target_files", [os.path.join(filepath, filename)]
            )
            self.api.llm.reload_rag(self.chatbot.target_files)
            self.api.llm.send_request(
                action=LLMActionType.RAG,
                prompt="Summarize the text and provide a synopsis of the content. Be concise and informative.",
                llm_request=LLMRequest.from_default(),
            )

    def show_layers(self):
        self.api.art.canvas.show_layers()

    def on_reset_paths_signal(self):
        self.reset_path_settings()

    def restart(self):
        # Save the current state
        self.save_state()

        # Close the main window
        self.close()

        # Start a new instance of the application
        QProcess.startDetached(sys.executable, sys.argv)

    def toggle_window_visibility(self):
        """Toggle window visibility and update the menu text."""
        if self.isVisible():
            self.hide()
            self.toggle_visibility_action.setText("Show Window")
        else:
            self.showNormal()
            self.activateWindow()  # Ensure window gets focus when showing
            self.toggle_visibility_action.setText("Hide Window")

        # Update the tray menu so it reflects the current state immediately
        self.tray_icon.setContextMenu(self.tray_menu)

    def handle_single_click(self):
        """Handle single-click on the tray icon."""
        print("Single click handler executing")
        # Create a dropdown menu
        menu = QMenu()

        # Dynamically set the text based on current window visibility
        show_hide_text = "Hide Window" if self.isVisible() else "Show Window"
        show_action = QAction(show_hide_text, self)
        quit_action = QAction("Quit", self)

        # Connect actions to their respective slots
        show_action.triggered.connect(self.toggle_window_visibility)
        quit_action.triggered.connect(QApplication.quit)

        # Add actions to the menu
        menu.addAction(show_action)
        menu.addAction(quit_action)

        # Display the menu under the tray icon
        menu.exec(QCursor.pos())

    def handle_double_click(self):
        """Handle double-click on the tray icon."""
        print("Double click detected")
        self.toggle_window_visibility()  # Use toggle instead of just showing

    @staticmethod
    def on_write_file_signal(data: Dict):
        """
        Writes data to a file.
        :param data: Dict
        :return: None
        """
        args = data["args"]
        if len(args) == 1:
            message = args[0]
        else:
            message = args[1]
        with open("output.txt", "w") as f:
            f.write(message)

    @staticmethod
    def on_bash_execute_signal(data: Dict) -> str:
        """
        Takes a message from the LLM and strips bash commands from it.
        Passes bash command to the bash_execute function.
        :param data: Dict
        :return:
        """
        args = data["args"]
        return bash_execute(args[0])

    def on_theme_changed_signal(self, data: Dict):
        self.set_stylesheet(
            data.get("dark_mode", False),
            data.get("override_system_theme", False),
        )
        self.update_icons()

    def update_icons(self):
        theme = (
            "dark" if self.application_settings.dark_mode_enabled else "light"
        )
        self.icon_manager.update_icons(theme)

    def initialize_ui(self):
        self.logger.debug("Loading UI")

        self.ui.setupUi(self)

        self.icon_manager = IconManager(self.icons, self.ui)

        if not AIRUNNER_ART_ENABLED:
            self._disable_aiart_gui_elements()

        active_index = 0
        tab = Tab.objects.filter_by(section="center", active=True)
        if tab:
            print(tab)
            active_index = tab[0].index

        self.ui.center_tab_container.setCurrentIndex(active_index)
        # on tab change, track the Tab index on the Tab object
        self.ui.center_tab_container.currentChanged.connect(
            lambda index: self.update_tab_index(section="center", index=index)
        )

        self.set_stylesheet()
        self.restore_state()
        # Configure default splitter sizes to maximize the canvas area (index 1)
        default_splitter_config = {
            "main_window_splitter": {
                "index_to_maximize": 1,
                "min_other_size": 50,
            }
        }
        load_splitter_settings(
            self.ui,
            ["main_window_splitter"],
            default_maximize_config=default_splitter_config,
        )

        self.status_widget = StatusWidget()
        self.statusBar().addPermanentWidget(self.status_widget)
        self.api.clear_status_message()
        self.initialize_widget_elements()
        self._load_plugins()
        self.api.main_window_loaded(self)

        if not AIRUNNER_DISCORD_URL:
            self.ui.actionDiscord.deleteLater()

        self._toggle_agent_workflow_feature(self.enable_workflows)

    def update_tab_index(self, section: str, index: int):
        Tab.objects.update_by(filter=dict(section=section), active=False)
        Tab.objects.update_by(
            filter=dict(section=section, index=index), active=True
        )

    def _disable_aiart_gui_elements(self):
        self.ui.center_widget.hide()
        self.ui.menuImage.hide()
        self.ui.menuFilters.hide()
        self.ui.menuStable_Diffusion.hide()
        self.ui.menuArt.hide()

        self.ui.center_widget.deleteLater()
        self.ui.actionToggle_Stable_Diffusion.deleteLater()
        self.ui.actionToggle_Controlnet.deleteLater()
        self.ui.menuImage.deleteLater()
        self.ui.menuFilters.deleteLater()
        self.ui.menuStable_Diffusion.deleteLater()
        self.ui.actionBrowse_AI_Runner_Path.deleteLater()
        self.ui.actionBrowse_Images_Path_2.deleteLater()
        self.ui.actionArt.deleteLater()
        self.ui.menuArt.deleteLater()
        self.ui.actionCut.deleteLater()
        self.ui.actionCopy.deleteLater()
        self.ui.actionPaste.deleteLater()
        self.ui.actionRotate_90_clockwise.deleteLater()
        self.ui.actionRotate_90_counterclockwise.deleteLater()
        self.ui.actionPrompt_Browser.deleteLater()

    def _load_plugins(self):
        base_path = self.path_settings.base_path
        path = os.path.join(base_path, "plugins")
        self.logger.info(f"Loading pluggins from {path}")
        plugin_loader = PluginLoader(plugin_dir=path)
        plugins = plugin_loader.load_plugins()

        if len(plugins) == 0:
            self.logger.info("No plugins found at " + path)
        else:
            self.logger.info("Loading plugins from " + path)
            for plugin in plugins:
                if hasattr(plugin, "get_widget"):
                    widget = plugin.get_widget()
                    self.ui.center_tab_container.addTab(widget, plugin.name)

    def initialize_widget_elements(self):
        for item in (
            (self.ui.actionToggle_LLM, self.application_settings.llm_enabled),
            (
                self.ui.actionToggle_Text_to_Speech,
                self.application_settings.tts_enabled,
            ),
            (
                self.ui.actionToggle_Speech_to_Text,
                self.application_settings.stt_enabled,
            ),
            (
                self.ui.actionToggle_Stable_Diffusion,
                self.application_settings.sd_enabled,
            ),
            (
                self.ui.actionToggle_Controlnet,
                self.application_settings.controlnet_enabled,
            ),
        ):
            item[0].blockSignals(True)
            item[0].setChecked(item[1] or False)
            item[0].blockSignals(False)
        self.initialized = True

    def layer_opacity_changed(self, _attr_name, value=None, _widget=None):
        self.api.art.canvas.layer_opacity_changed(value)

    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        for v in self.shortcut_keys:
            if v.key == event.key():
                for signal in SignalCode:
                    if signal.value == v.signal:
                        self.emit_signal(signal)
                        break

    def key_text(self, key_name):
        for shortcutkey in self.shortcut_keys:
            if shortcutkey.name == key_name:
                return shortcutkey.text
        return ""

    def on_save_stablediffusion_prompt_signal(self, data: Dict):
        self.create_saved_prompt(
            {
                "prompt": data["prompt"],
                "negative_prompt": data["negative_prompt"],
                "secondary_prompt": data["secondary_prompt"],
                "secondary_negative_prompt": data["secondary_negative_prompt"],
            }
        )

    def set_path_settings(self, key, val):
        self.update_path_settings(key, val)

    def on_nsfw_content_detected_signal(self):
        # display message in status
        self.api.application_error(AIRUNNER_NSFW_CONTENT_DETECTED_MESSAGE)

    def closeEvent(self, event):
        event.ignore()
        self.handle_close()

    def handle_close(self):
        """Override close to minimize to tray instead of exiting."""
        self.quit()

    def quit(self):
        self.logger.debug("Quitting")
        self.save_state()
        self.api.quit_application()

    def handle_quit_application_signal(self):
        self.hide()
        QApplication.quit()
        self.close()

    def show_settings_path(self, name, default_path=None):
        path = getattr(self.path_settings, name)
        show_path(default_path if default_path and path == "" else path)

    def set_icons(self, icon_name, widget_name, theme):
        if not self.initialized:
            return

        icon = QtGui.QIcon()
        icon.addPixmap(
            QtGui.QPixmap(f":/{theme}/icons/feather/{theme}/{icon_name}.svg"),
            QtGui.QIcon.Mode.Normal,
            QtGui.QIcon.State.Off,
        )
        getattr(self.ui, widget_name).setIcon(icon)
        self.update()

    def toggle_nsfw_filter(self):
        self.set_nsfw_filter_tooltip()

    def set_nsfw_filter_tooltip(self):
        self.ui.actionSafety_Checker.setToolTip(
            f"Click to {'enable' if not self.application_settings.nsfw_filter else 'disable'} NSFW filter"
        )

    def on_toggle_fullscreen_signal(self):
        if self.isFullScreen():
            self.showNormal()
        else:
            self.showFullScreen()

    def on_unload_non_sd_models(self, data: Dict = None):
        if not self.memory_settings.prevent_unload_on_llm_image_generation:
            sd_device = (
                AIRUNNER_MEM_SD_DEVICE or self.memory_settings.default_gpu_sd
            )
            llm_device = (
                AIRUNNER_MEM_LLM_DEVICE or self.memory_settings.default_gpu_llm
            )
            tts_device = (
                AIRUNNER_MEM_TTS_DEVICE or self.memory_settings.default_gpu_tts
            )
            stt_device = (
                AIRUNNER_MEM_STT_DEVICE or self.memory_settings.default_gpu_stt
            )

            if sd_device == tts_device:
                self._tts_generator_worker.unload()

            if sd_device == stt_device:
                self._stt_audio_processor_worker.unload()

        callback = data.get("callback", None)
        if callback:
            callback()

    def on_load_non_sd_models(self, data: Dict = None):
        if self.application_settings.llm_enabled:
            self._llm_generate_worker.load()
        if self.application_settings.tts_enabled:
            self._tts_generator_worker.load()
        if self.application_settings.stt_enabled:
            self._stt_audio_processor_worker.load()
        callback = data.get("callback", None)
        if callback:
            callback(data)

    def on_toggle_llm(self, data: Dict = None, val=None):
        if val is None:
            val = not self.application_settings.llm_enabled
        self._update_action_button(
            ModelType.LLM,
            self.ui.actionToggle_LLM,
            val,
            SignalCode.LLM_LOAD_SIGNAL,
            SignalCode.LLM_UNLOAD_SIGNAL,
            "llm_enabled",
            data,
        )

    def on_toggle_sd(self, data: Dict):
        val = data.get("enabled", False)
        self._update_action_button(
            ModelType.SD,
            self.ui.actionToggle_Stable_Diffusion,
            val,
            SignalCode.SD_LOAD_SIGNAL,
            SignalCode.SD_UNLOAD_SIGNAL,
            "sd_enabled",
            data,
        )

    def on_toggle_tts(self, data: Dict = None, val=None):
        if val is None:
            val = data.get(
                "enabled", not self.application_settings.tts_enabled
            )
        self._update_action_button(
            ModelType.TTS,
            self.ui.actionToggle_Text_to_Speech,
            val,
            SignalCode.TTS_ENABLE_SIGNAL,
            SignalCode.TTS_DISABLE_SIGNAL,
            "tts_enabled",
            data,
        )

    def _update_action_button(
        self,
        model_type,
        element,
        val: bool,
        load_signal: SignalCode,
        unload_signal: SignalCode,
        application_setting: str = None,
        data: Dict = None,
    ):
        if self._model_status[model_type] is ModelStatus.LOADING:
            val = not val
        element.blockSignals(True)
        element.setChecked(val)
        element.blockSignals(False)
        QApplication.processEvents()
        if application_setting:
            self.update_application_settings(application_setting, val)
        if self._model_status[model_type] is not ModelStatus.LOADING:
            if model_type is ModelType.TTS:
                if val:
                    self.api.tts.start()
                else:
                    self.api.tts.stop()
            else:
                if val:
                    self.emit_signal(load_signal, data)
                else:
                    self.emit_signal(unload_signal, data)

    def save_state(self):
        if self.quitting:
            return
        self.quitting = True
        self.logger.debug("Saving window state")

        self.qsettings.beginGroup("window_settings")
        self.qsettings.setValue("is_maximized", self.isMaximized())
        self.qsettings.setValue("is_fullscreen", self.isFullScreen())
        self.qsettings.setValue("width", self.width())
        self.qsettings.setValue("height", self.height())
        self.qsettings.setValue("x_pos", self.pos().x())
        self.qsettings.setValue("y_pos", self.pos().y())
        self.qsettings.setValue(
            "mode_tab_widget_index",
            self.ui.generator_widget.ui.generator_form_tabs.currentIndex(),
        )
        self.qsettings.endGroup()
        save_splitter_settings(self.ui, ["main_window_splitter"])

    def restore_state(self):
        """
        Restore the window based on the previous state.
        """
        self.logger.debug("Restoring state")

        # Get the window settings
        settings = self.window_settings

        # Resize the window
        self.setMinimumSize(512, 512)
        width = int(settings["width"])
        height = int(settings["height"])
        self.resize(width, height)

        # Move the window
        x_pos = int(settings["x_pos"])
        y_pos = int(settings["y_pos"])
        self.move(x_pos, y_pos)

        # Show the window
        if settings is not None and settings.get("is_maximized", False):
            self.logger.info("Restoring window to maximized state")
            self.showMaximized()
        elif settings is not None and settings.get("is_fullscreen", False):
            self.logger.info("Restoring window to fullscreen state")
            self.showFullScreen()
        else:
            self.logger.info("Restoring window to normal state")
            self.showNormal()

        # Raise the window to the top of the stack
        self.raise_()

    ##### End window properties #####
    #################################

    ###### Window handlers ######
    def on_toggle_tool_signal(self, data: Dict):
        self.toggle_tool(data["tool"], data["active"])

    def on_enable_workflows_toggled(self, message: Dict):
        self._toggle_agent_workflow_feature(message.get("enabled", False))

    def on_retranslate_ui_signal(self):
        self.ui.retranslateUi(self)

    def _toggle_agent_workflow_feature(self, enabled: bool):
        """
        Toggles the visibility of the workflow tab and menu based on the given value.
        """
        # --- Tab handling ---
        tab_widget = self.ui.center_tab_container
        if not self.workflow_tab:
            self.workflow_tab = self.ui.agent_workflow_tab

        if enabled:
            # Only add if not present
            if self.tab_backup != {}:
                tab_widget.addTab(
                    self.tab_backup["tab_widget"],
                    self.tab_backup["tab_text"],
                )
                self.tab_backup = {}
        else:
            # Remove if present
            index = tab_widget.indexOf(self.workflow_tab)
            self.tab_backup = dict(
                tab_text=tab_widget.tabText(index),
                tab_widget=tab_widget.widget(index),
            )
            if index != -1:
                tab_widget.removeTab(index)

        # --- Menu handling ---
        # If menuWorkflow is a QMenu, use menuBar(). If it's an action, use setVisible.
        if hasattr(self.ui, "menuWorkflow"):
            menu = self.ui.menuWorkflow
            # Try both methods for robustness
            try:
                menu.menuAction().setVisible(enabled)
            except Exception:
                try:
                    menu.setVisible(enabled)
                except Exception:
                    pass

    def show_nsfw_warning_popup(self):
        if self.application_settings.show_nsfw_warning:
            """
            Display a popup window which asks the user if they are sure they want to disable the NSFW filter
            along with a checkbox that allows the user to disable the warning in the future.
            """
            msg_box = QMessageBox()
            msg_box.setWindowTitle("Disable Safety Checker Warning")
            msg_box.setText(
                (
                    "WARNING\n\n"
                    "You are attempting to disable the safety checker (NSFW filter).\n"
                    "It is strongly recommended that you keep this enabled at all times.\n"
                    "The Safety Checker prevents potentially harmful content from being displayed.\n"
                    "Only disable it if you are sure the Image model you are using is not capable of generating "
                    "harmful content.\n"
                    "Disabling the safety checker is intended as a last resort for continual false positives and as a "
                    "research feature.\n"
                    "\n\n"
                    "Are you sure you want to disable the filter?"
                )
            )
            msg_box.setStandardButtons(
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
            )
            msg_box.setDefaultButton(QMessageBox.StandardButton.No)

            # Create a QCheckBox
            checkbox = QCheckBox("Do not show this warning again")
            # Add the checkbox to the message box
            msg_box.setCheckBox(checkbox)

            result = msg_box.exec()

            if result == QMessageBox.StandardButton.Yes:
                self._disable_nsfw_filter(not checkbox.isChecked())

            self.ui.actionSafety_Checker.blockSignals(True)
            self.ui.actionSafety_Checker.setChecked(
                self.application_settings.nsfw_filter
            )
            self.ui.actionSafety_Checker.blockSignals(False)
        else:
            self._disable_nsfw_filter()

    def _disable_nsfw_filter(self, show_nsfw_warning=None):
        """
        Do not call this function directly.
        :return:
        """
        # User confirmed to disable the NSFW filter
        # Update the settings accordingly
        self.update_application_settings("nsfw_filter", False)
        # Update the show_nsfw_warning setting based on the checkbox state
        if show_nsfw_warning is not None:
            self.update_application_settings(
                "show_nsfw_warning", show_nsfw_warning
            )
        self.toggle_nsfw_filter()
        self.api.art.unload_safety_checker()

    ###### End window handlers ######

    def show_update_message(self):
        self.api.application_status(
            f"New version available: {self.latest_version}"
        )

    def show_update_popup(self):
        self.update_popup = UpdateWindow()

    @staticmethod
    def show_setup_wizard():
        AppInstaller(close_on_cancel=False)

    def showEvent(self, event):
        """Override to update the tray menu text when window is shown."""
        super().showEvent(event)
        # Make sure we update the menu text whenever the window is shown
        if hasattr(self, "toggle_visibility_action"):
            self.toggle_visibility_action.setText("Hide Window")
            # Update the tray menu so it reflects the current state immediately
            self.tray_icon.setContextMenu(self.tray_menu)

        self.logger.debug("showEvent called, initializing window")
        self._initialize_window()
        self._initialize_default_buttons()
        self._initialize_filter_actions()
        self.initialized = True
        self.update_icons()
        self.logger.debug("Showing window")
        self._set_keyboard_shortcuts()

    def on_keyboard_shortcuts_updated(self):
        self._set_keyboard_shortcuts()

    def _set_keyboard_shortcuts(self):
        quit_key = ShortcutKeys.objects.filter_by_first(display_name="Quit")
        brush_key = ShortcutKeys.objects.filter_by_first(display_name="Brush")
        eraser_key = ShortcutKeys.objects.filter_by_first(
            display_name="Eraser"
        )
        move_tool_key = ShortcutKeys.objects.filter_by_first(
            display_name="Move Tool"
        )

        if quit_key is not None:
            key_sequence = QKeySequence(quit_key.key | quit_key.modifiers)
            self.ui.actionQuit.setShortcut(key_sequence)
            self.ui.actionQuit.setToolTip(
                f"{quit_key.display_name} ({quit_key.text})"
            )

        if brush_key is not None:
            key_sequence = QKeySequence(brush_key.key | brush_key.modifiers)
            self.ui.actionToggle_Brush.setShortcut(key_sequence)
            self.ui.actionToggle_Brush.setToolTip(
                f"{brush_key.display_name} ({brush_key.text})"
            )

        if eraser_key is not None:
            key_sequence = QKeySequence(eraser_key.key | eraser_key.modifiers)
            self.ui.actionToggle_Eraser.setShortcut(key_sequence)
            self.ui.actionToggle_Eraser.setToolTip(
                f"{eraser_key.display_name} ({eraser_key.text})"
            )

        if move_tool_key is not None:
            key_sequence = QKeySequence(
                move_tool_key.key | move_tool_key.modifiers
            )
            self.ui.actionToggle_Active_Grid_Area.setShortcut(key_sequence)
            self.ui.actionToggle_Active_Grid_Area.setToolTip(
                f"{move_tool_key.display_name} ({move_tool_key.text})"
            )

    def _initialize_workers(self):
        self.logger.debug("Initializing worker manager")
        self.logger.info("imported workers, initializing")
        self._mask_generator_worker = create_worker(MaskGeneratorWorker)
        self._sd_worker = create_worker(SDWorker)
        if AudioCaptureWorker is not None:
            self._stt_audio_capture_worker = create_worker(AudioCaptureWorker)
        if AudioProcessorWorker is not None:
            self._stt_audio_processor_worker = create_worker(
                AudioProcessorWorker
            )
        if TTSGeneratorWorker is not None:
            self._tts_generator_worker = create_worker(TTSGeneratorWorker)
        if TTSVocalizerWorker is not None:
            self._tts_vocalizer_worker = create_worker(TTSVocalizerWorker)
        self._llm_generate_worker = create_worker(LLMGenerateWorker)

        self.logger.info("INITIALIZE WORKERS COMPLETE")

    def _initialize_filter_actions(self):
        image_filters = ImageFilter.objects.all()
        try:
            for image_filter in image_filters:
                action = self.ui.menuFilters.addAction(
                    image_filter.display_name
                )
                action.triggered.connect(
                    partial(self.display_filter_window, image_filter)
                )
        except RuntimeError as e:
            if AIRUNNER_ART_ENABLED:
                self.logger.warning(f"Error setting SD status text: {e}")

    @staticmethod
    def display_filter_window(image_filter):
        FilterWindow(image_filter.id)

    def _initialize_default_buttons(self):
        self.ui.actionSafety_Checker.blockSignals(True)
        self.ui.actionSafety_Checker.setChecked(
            self.application_settings.nsfw_filter
        )
        self.ui.actionSafety_Checker.blockSignals(False)

    def _initialize_window(self):
        # self.center()
        screen = QGuiApplication.primaryScreen()  # Use primaryScreen

        if not screen:
            self.logger.warning(
                "Could not get primary screen. Falling back to default size."
            )
            # Fallback to a default size if screen info is unavailable
            default_width, default_height = 1024, 768
            self.resize(default_width, default_height)
            self.setMinimumSize(512, 512)
            # Set maximum size to something reasonable if screen info is missing
            self.setMaximumSize(
                default_width, default_height
            )  # Or a larger sensible max
        else:
            screen_geometry = screen.availableGeometry()
            self.logger.info(
                f"Available screen geometry: "
                f"x={screen_geometry.x()}, y={screen_geometry.y()}, "
                f"width={screen_geometry.width()}, height={screen_geometry.height()}"
            )

            # Set geometry using explicit move and resize
            self.move(screen_geometry.topLeft())
            self.resize(screen_geometry.size())

            self.setMinimumSize(512, 512)
            # Ensure maximum size is at least the minimum size and not smaller than the available geometry
            max_width = max(screen_geometry.width(), self.minimumWidth())
            max_height = max(screen_geometry.height(), self.minimumHeight())
            self.setMaximumSize(max_width, max_height)

        self.setWindowIcon(
            QIcon(
                os.path.join(self.path_settings.base_path, "images/icon.png")
            )
        )
        self.set_window_title()

    def center(self):
        available_geometry = (
            QGuiApplication.primaryScreen().availableGeometry()
        )
        frame_geometry = self.frameGeometry()
        frame_geometry.moveCenter(available_geometry.center())
        self.move(frame_geometry.topLeft())

    def set_window_title(self):
        """
        Overrides base method to set the window title
        :return:
        """
        self.setWindowTitle(self._window_title)

    def handle_unknown(self, message):
        self.logger.error(f"Unknown message code: {message}")

    def clear_all_prompts(self):
        self.prompt = ""
        self.negative_prompt = ""
        self.api.clear_prompts()

    def new_batch(self, index, image, data):
        self.generator_tab_widget.new_batch(index, image, data)

    def on_model_status_changed_signal(self, data):
        model = data["model"]
        status = data["status"]
        if self._model_status[model] is status:
            return
        self._model_status[model] = status
        if model is ModelType.SD:
            self.ui.actionToggle_Stable_Diffusion.setDisabled(
                status is ModelStatus.LOADING
            )
            if status is ModelStatus.FAILED:
                self.ui.actionToggle_Stable_Diffusion.setChecked(False)
        elif model is ModelType.CONTROLNET:
            self.ui.actionToggle_Controlnet.setDisabled(
                status is ModelStatus.LOADING
            )
            if status is ModelStatus.FAILED:
                self.ui.actionToggle_Controlnet.setChecked(False)
        elif model is ModelType.LLM:
            self.ui.actionToggle_LLM.setDisabled(status is ModelStatus.LOADING)
            if status is ModelStatus.FAILED:
                self.logger.warning("LLM failed to load")
                self.ui.actionToggle_LLM.setChecked(False)
        elif model is ModelType.TTS:
            self.ui.actionToggle_Text_to_Speech.setDisabled(
                status is ModelStatus.LOADING
            )
            if status is ModelStatus.FAILED:
                self.ui.actionToggle_Text_to_Speech.setChecked(False)
        elif model is ModelType.STT:
            self.ui.actionToggle_Speech_to_Text.setDisabled(
                status is ModelStatus.LOADING
            )
            if status is ModelStatus.FAILED:
                self.ui.actionToggle_Speech_to_Text.setChecked(False)
        self.initialize_widget_elements()
        QApplication.processEvents()

    def _generate_drawingpad_mask(self):
        width = self.application_settings.working_width
        height = self.application_settings.working_height
        img = Image.new("RGB", (width, height), (0, 0, 0))
        base64_image = convert_image_to_binary(img)

        drawing_pad_settings = DrawingPadSettings.objects.first()
        drawing_pad_settings.mask = base64_image
        drawing_pad_settings.save()

    def display_missing_models_error(self, data):
        msg_box = QMessageBox()
        msg_box.setIcon(QMessageBox.Critical)
        msg_box.setWindowTitle(data.get("title", "Error: Missing models"))
        msg_box.setText(data.get("message", "Something went wrong"))
        msg_box.exec()



from PySide6.QtWidgets import (
    QLineEdit,
    QDialog,
    QFormLayout,
    QDialogButtonBox,
)


class AddPortDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Add Port")
        layout = QFormLayout(self)

        self.port_name_input = QLineEdit(self)
        self.port_type_input = QLineEdit(
            self
        )  # Simple text for now, could be dropdown

        layout.addRow("Port Name:", self.port_name_input)
        layout.addRow("Port Type (optional):", self.port_type_input)

        self.buttons = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel, self
        )
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)
        layout.addWidget(self.buttons)

    def get_port_info(self):
        return self.port_name_input.text(), self.port_type_input.text()



import unittest
from unittest.mock import patch, MagicMock
from PySide6.QtWidgets import QWidget, QApplication
from airunner.api import API
from airunner.enums import SignalCode, LLMActionType
from airunner.handlers.llm import LLMRequest, LLMResponse
from airunner.handlers.stablediffusion.image_request import ImageRequest


class TestAPI(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Create a real QApplication instance once for the entire test suite
        cls.qapp = QApplication([])

    @classmethod
    def tearDownClass(cls):
        # Clean up the QApplication instance after all tests
        cls.qapp.quit()

    def setUp(self):
        # Pass initialize_app=False to avoid starting the full application
        self.api = API(initialize_app=False, initialize_gui=False)
        # Mock the app and main_window attributes with a QWidget
        self.api.app = MagicMock()
        self.api.app.main_window = QWidget()

    @patch("airunner.api.create_worker")
    @patch("airunner.api.setup_database")
    def test_initialization(self, mock_setup_database, mock_create_worker):
        mock_worker = MagicMock()
        mock_create_worker.return_value = mock_worker

        api = API(initialize_app=True, initialize_gui=False)

        mock_setup_database.assert_called_once()
        mock_create_worker.assert_called_once()
        mock_worker.add_to_queue.assert_called_with("scan_for_models")

    @patch("airunner.api.API.emit_signal")
    def test_send_llm_request(self, mock_emit_signal):
        prompt = "Test prompt"
        llm_request = LLMRequest()
        action = LLMActionType.CHAT

        self.api.llm.send_request(
            prompt, llm_request, action, do_tts_reply=True
        )

        mock_emit_signal.assert_called_once_with(
            SignalCode.LLM_TEXT_GENERATE_REQUEST_SIGNAL,
            {
                "llm_request": True,
                "request_data": {
                    "action": action,
                    "prompt": prompt,
                    "llm_request": llm_request,
                    "do_tts_reply": True,
                },
            },
        )

    @patch("airunner.api.API.emit_signal")
    def test_send_tts_request(self, mock_emit_signal):
        response = LLMResponse()

        self.api.llm.send_llm_text_streamed_signal(response)

        mock_emit_signal.assert_called_once_with(
            SignalCode.LLM_TEXT_STREAMED_SIGNAL, {"response": response}
        )

    @patch("airunner.api.API.emit_signal")
    def test_send_image_request(self, mock_emit_signal):
        image_request = ImageRequest()

        self.api.send_image_request(image_request)

        mock_emit_signal.assert_called_once_with(
            SignalCode.DO_GENERATE_SIGNAL, {"image_request": image_request}
        )

    @patch("airunner.api.render_ui_from_spec", autospec=True)
    @patch("PySide6.QtWidgets.QDialog.exec", autospec=True)
    def test_show_hello_world_window(self, mock_exec, mock_render_ui):
        self.api.show_hello_world_window()

        # Debugging: Print call arguments to verify invocation
        print("render_ui_from_spec call args:", mock_render_ui.call_args_list)
        print("QDialog.exec call args:", mock_exec.call_args_list)

        mock_render_ui.assert_called_once()
        mock_exec.assert_called_once()

    @patch("airunner.api.load_ui_file", autospec=True)
    @patch("PySide6.QtWidgets.QDialog.exec", autospec=True)
    def test_show_dynamic_ui(self, mock_exec, mock_load_ui):
        # Return a QWidget dynamically to ensure QApplication is initialized
        mock_load_ui.return_value = QWidget()

        ui_file_path = "test.ui"

        self.api.show_dynamic_ui(ui_file_path)

        # Debugging: Print call arguments to verify invocation
        print("load_ui_file call args:", mock_load_ui.call_args_list)
        print("QDialog.exec call args:", mock_exec.call_args_list)

        mock_load_ui.assert_called_once_with(ui_file_path, unittest.mock.ANY)
        mock_exec.assert_called_once()

    @patch("airunner.api.load_ui_from_string", autospec=True)
    @patch("PySide6.QtWidgets.QDialog.exec", autospec=True)
    def test_show_dynamic_ui_from_string(self, mock_exec, mock_load_ui):
        # Return a QWidget dynamically to ensure QApplication is initialized
        mock_load_ui.return_value = QWidget()

        ui_content = "<ui></ui>"
        data = {"ui_content": ui_content}

        self.api.show_dynamic_ui_from_string(data)

        # Debugging: Print call arguments to verify invocation
        print("load_ui_from_string call args:", mock_load_ui.call_args_list)
        print("QDialog.exec call args:", mock_exec.call_args_list)

        mock_load_ui.assert_called_once_with(
            ui_content, unittest.mock.ANY, unittest.mock.ANY
        )
        mock_exec.assert_called_once()


if __name__ == "__main__":
    unittest.main()



from airunner.data.models.voice_settings import VoiceSettings
from airunner.data.models.sound_settings import SoundSettings
from airunner.data.models.openvoice_settings import OpenVoiceSettings
from airunner.data.models.active_grid_settings import ActiveGridSettings
from airunner.data.models.application_settings import ApplicationSettings
from airunner.data.models.controlnet_settings import ControlnetSettings
from airunner.data.models.image_to_image_settings import ImageToImageSettings
from airunner.data.models.outpaint_settings import OutpaintSettings
from airunner.data.models.chatstore import Chatstore
from airunner.data.models.drawingpad_settings import DrawingPadSettings
from airunner.data.models.metadata_settings import MetadataSettings
from airunner.data.models.generator_settings import GeneratorSettings
from airunner.data.models.llm_generator_settings import LLMGeneratorSettings
from airunner.data.models.speech_t5_settings import SpeechT5Settings
from airunner.data.models.espeak_settings import EspeakSettings
from airunner.data.models.stt_settings import STTSettings
from airunner.data.models.schedulers import Schedulers
from airunner.data.models.brush_settings import BrushSettings
from airunner.data.models.grid_settings import GridSettings
from airunner.data.models.path_settings import PathSettings
from airunner.data.models.memory_settings import MemorySettings
from airunner.data.models.chatbot import Chatbot
from airunner.data.models.user import User
from airunner.data.models.target_files import TargetFiles
from airunner.data.models.target_directories import TargetDirectories
from airunner.data.models.ai_models import AIModels
from airunner.data.models.shortcut_keys import ShortcutKeys
from airunner.data.models.lora import Lora
from airunner.data.models.saved_prompt import SavedPrompt
from airunner.data.models.embedding import Embedding
from airunner.data.models.prompt_template import PromptTemplate
from airunner.data.models.controlnet_model import ControlnetModel
from airunner.data.models.font_setting import FontSetting
from airunner.data.models.pipeline_model import PipelineModel
from airunner.data.models.conversation import Conversation
from airunner.data.models.summary import Summary
from airunner.data.models.image_filter import ImageFilter
from airunner.data.models.image_filter_value import ImageFilterValue
from airunner.data.models.whisper_settings import WhisperSettings
from airunner.data.models.news import RSSFeed, Category, Article
from airunner.data.models.tab import Tab
from airunner.data.models.workflow import Workflow
from airunner.data.models.workflow_node import WorkflowNode
from airunner.data.models.workflow_connection import WorkflowConnection
from airunner.data.models.rag_settings import RAGSettings
from airunner.data.models.language_settings import LanguageSettings
from airunner.data.models.base import Base


classes = [
    ActiveGridSettings,
    ApplicationSettings,
    ControlnetSettings,
    ImageToImageSettings,
    OutpaintSettings,
    Chatstore,
    DrawingPadSettings,
    MetadataSettings,
    GeneratorSettings,
    LLMGeneratorSettings,
    SpeechT5Settings,
    EspeakSettings,
    STTSettings,
    Schedulers,
    BrushSettings,
    GridSettings,
    PathSettings,
    MemorySettings,
    Chatbot,
    User,
    TargetFiles,
    TargetDirectories,
    AIModels,
    ShortcutKeys,
    Lora,
    SavedPrompt,
    Embedding,
    PromptTemplate,
    ControlnetModel,
    FontSetting,
    PipelineModel,
    Conversation,
    Summary,
    ImageFilter,
    ImageFilterValue,
    WhisperSettings,
    RSSFeed,
    Category,
    Article,
    Tab,
    VoiceSettings,
    OpenVoiceSettings,
    Workflow,
    WorkflowNode,
    WorkflowConnection,
    RAGSettings,
    LanguageSettings,
    Base,
]

class_names = []
table_to_class = {}
for cls in classes:
    if cls is not Base:
        table_to_class[cls.__tablename__] = cls
    class_names.append(cls.__name__)

__all__ = class_names



from PySide6.QtWidgets import QApplication, QMainWindow, QVBoxLayout, QLabel, QWidget, QDialog

def render_ui_from_spec(spec, parent_window):
    if spec["type"] == "window":
        if isinstance(parent_window, QDialog):
            layout = QVBoxLayout(parent_window)
        else:
            parent_window.setWindowTitle(spec.get("title", "Untitled"))
            central_widget = QWidget()
            layout = QVBoxLayout(central_widget)
            parent_window.setCentralWidget(central_widget)

        for widget_spec in spec.get("widgets", []):
            if widget_spec["type"] == "label":
                label = QLabel(widget_spec.get("text", ""))
                layout.addWidget(label)

        if isinstance(parent_window, QDialog):
            parent_window.setLayout(layout)
        parent_window.show()

def test_hello_world_window():
    app = QApplication([])
    main_window = QMainWindow()
    spec = {
        "type": "window",
        "title": "Hello Window",
        "layout": "vertical",
        "widgets": [
            {"type": "label", "text": "Hello, world!"}
        ]
    }
    render_ui_from_spec(spec, main_window)
    app.exec()


# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'slider.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractSpinBox, QApplication, QDoubleSpinBox, QGridLayout,
    QGroupBox, QSizePolicy, QSlider, QWidget)

class Ui_slider_widget(object):
    def setupUi(self, slider_widget):
        if not slider_widget.objectName():
            slider_widget.setObjectName(u"slider_widget")
        slider_widget.resize(548, 50)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(slider_widget.sizePolicy().hasHeightForWidth())
        slider_widget.setSizePolicy(sizePolicy)
        slider_widget.setMinimumSize(QSize(0, 0))
        slider_widget.setMaximumSize(QSize(16777215, 16777215))
        slider_widget.setStyleSheet(u"")
        self.gridLayout = QGridLayout(slider_widget)
        self.gridLayout.setSpacing(0)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.groupBox = QGroupBox(slider_widget)
        self.groupBox.setObjectName(u"groupBox")
        sizePolicy.setHeightForWidth(self.groupBox.sizePolicy().hasHeightForWidth())
        self.groupBox.setSizePolicy(sizePolicy)
        self.groupBox.setMinimumSize(QSize(0, 0))
        self.groupBox.setMaximumSize(QSize(16777215, 16777215))
        font = QFont()
        font.setPointSize(9)
        self.groupBox.setFont(font)
        self.gridLayout_2 = QGridLayout(self.groupBox)
        self.gridLayout_2.setSpacing(0)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.gridLayout_2.setContentsMargins(5, 5, 5, 5)
        self.slider = QSlider(self.groupBox)
        self.slider.setObjectName(u"slider")
        sizePolicy.setHeightForWidth(self.slider.sizePolicy().hasHeightForWidth())
        self.slider.setSizePolicy(sizePolicy)
        self.slider.setMinimumSize(QSize(0, 0))
        self.slider.setAutoFillBackground(False)
        self.slider.setStyleSheet(u"")
        self.slider.setMaximum(100)
        self.slider.setPageStep(1)
        self.slider.setOrientation(Qt.Orientation.Horizontal)

        self.gridLayout_2.addWidget(self.slider, 0, 0, 1, 1)

        self.slider_spinbox = QDoubleSpinBox(self.groupBox)
        self.slider_spinbox.setObjectName(u"slider_spinbox")
        sizePolicy.setHeightForWidth(self.slider_spinbox.sizePolicy().hasHeightForWidth())
        self.slider_spinbox.setSizePolicy(sizePolicy)
        self.slider_spinbox.setMinimumSize(QSize(0, 0))
        self.slider_spinbox.setMaximumSize(QSize(16777215, 16777215))
        self.slider_spinbox.setStyleSheet(u"")
        self.slider_spinbox.setWrapping(False)
        self.slider_spinbox.setFrame(False)
        self.slider_spinbox.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.slider_spinbox.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.UpDownArrows)
        self.slider_spinbox.setKeyboardTracking(True)
        self.slider_spinbox.setProperty(u"showGroupSeparator", False)
        self.slider_spinbox.setMaximum(1.000000000000000)
        self.slider_spinbox.setSingleStep(0.010000000000000)

        self.gridLayout_2.addWidget(self.slider_spinbox, 0, 1, 1, 1)


        self.gridLayout.addWidget(self.groupBox, 0, 0, 1, 1)


        self.retranslateUi(slider_widget)

        QMetaObject.connectSlotsByName(slider_widget)
    # setupUi

    def retranslateUi(self, slider_widget):
        slider_widget.setWindowTitle(QCoreApplication.translate("slider_widget", u"Form", None))
        self.groupBox.setTitle(QCoreApplication.translate("slider_widget", u"GroupBox", None))
    # retranslateUi




import os
from typing import (
    Any,
    List,
    Optional,
    Union,
    Dict,
    Type,
    Annotated,
)
import datetime
import platform

from llama_index.core.tools import BaseTool, FunctionTool
from llama_index.core.chat_engine.types import AgentChatResponse
from llama_index.core.base.llms.types import ChatMessage
from llama_index.core.memory import BaseMemory
from llama_index.core.llms.llm import LLM
from llama_index.core.storage.chat_store.base import BaseChatStore
from llama_index.core.storage.chat_store import SimpleChatStore

from transformers import AutoModelForCausalLM, AutoTokenizer

from airunner.enums import (
    LANGUAGE_DISPLAY_MAP,
    AvailableLanguage,
    GeneratorSection,
    LLMActionType,
    SignalCode,
    ImagePreset,
)
from airunner.data.models import Conversation, User, Tab
from airunner.utils.application.mediator_mixin import MediatorMixin
from airunner.gui.windows.main.settings_mixin import SettingsMixin
from airunner.handlers.llm.agent import (
    RAGMixin,
    ExternalConditionStoppingCriteria,
)
from airunner.handlers.llm.agent.tools import ChatEngineTool, ReActAgentTool
from airunner.handlers.llm.agent.chat_engine import RefreshSimpleChatEngine
from airunner.handlers.llm.agent import WeatherMixin
from airunner.handlers.llm.agent.memory import ChatMemoryBuffer
from airunner.handlers.llm.storage.chat_store import DatabaseChatStore
from airunner.handlers.llm.llm_request import LLMRequest
from airunner.handlers.llm.llm_response import LLMResponse
from airunner.handlers.llm.llm_settings import LLMSettings
from airunner.handlers.llm import HuggingFaceLLM
from airunner.data.models import Conversation
from airunner.settings import (
    AIRUNNER_LLM_CHAT_STORE,
    AIRUNNER_ART_ENABLED,
    AIRUNNER_MOOD_PROMPT_OVERRIDE,
)
from airunner.utils.llm.language import detect_language


class BaseAgent(
    MediatorMixin,
    SettingsMixin,
    RAGMixin,
    WeatherMixin,
):
    """
    Base class for all agents.

    Args:
        default_tool_choice (Optional[Union[str, dict], optional): The default tool choice. Defaults to None.
        llm_settings (LLMSettings, optional): The LLM settings. Defaults to LLMSettings().
    """

    def __init__(
        self,
        default_tool_choice: Optional[Union[str, dict]] = None,
        llm_settings: LLMSettings = LLMSettings(),
        use_memory: bool = True,
        *args,
        **kwargs,
    ) -> None:
        self.default_tool_choice: Optional[Union[str, dict]] = (
            default_tool_choice
        )
        self._prompt = None
        self._language = None
        self._llm_request: Optional[LLMRequest] = None
        self.llm_settings: LLMSettings = llm_settings
        self._use_memory: bool = use_memory
        self._action: LLMActionType = LLMActionType.NONE
        self._chat_prompt: str = ""
        self._current_tab: Optional[Tab] = None
        self._streaming_stopping_criteria: Optional[
            ExternalConditionStoppingCriteria
        ] = None
        self._do_interrupt: bool = False
        self._llm: Optional[LLM] = None
        self._conversation: Optional[Conversation] = None
        self._conversation_id: Optional[int] = None
        self._user: Optional[User] = None
        self._chat_engine: Optional[RefreshSimpleChatEngine] = None
        self._mood_engine: Optional[RefreshSimpleChatEngine] = None
        self._summary_engine: Optional[RefreshSimpleChatEngine] = None
        self._chat_engine_tool: Optional[ChatEngineTool] = None
        self._mood_engine_tool: Optional[ChatEngineTool] = None
        self._update_user_data_engine = None
        self._update_user_data_tool = None
        self._summary_engine_tool: Optional[ChatEngineTool] = None
        self._information_scraper_tool: Optional[ChatEngineTool] = None
        self._information_scraper_engine: Optional[RefreshSimpleChatEngine] = (
            None
        )
        self._chat_store: Optional[Type[BaseChatStore]] = None
        self._chat_memory: Optional[ChatMemoryBuffer] = None
        self._memory: Optional[BaseMemory] = None
        self._react_tool_agent: Optional[ReActAgentTool] = None
        self._complete_response: str = ""
        self._store_user_tool: Optional[FunctionTool] = None
        self._webpage_html: str = ""
        self.model: Optional[AutoModelForCausalLM] = None
        self.tokenizer: Optional[AutoTokenizer] = None

        self.signal_handlers.update(
            {
                SignalCode.DELETE_MESSAGES_AFTER_ID: self.on_delete_messages_after_id
            }
        )
        super().__init__(*args, **kwargs)

    @property
    def prompt(self) -> str:
        return self._prompt

    @prompt.setter
    def prompt(self, value: str):
        self._prompt = value

    @property
    def language(self) -> str:
        bot_lang = AvailableLanguage(self.language_settings.bot_language)
        if bot_lang is AvailableLanguage.AUTO:
            bot_lang = detect_language(self.prompt)
        return LANGUAGE_DISPLAY_MAP.get(bot_lang)

    @language.setter
    def language(self, value: str):
        self._language = value

    @property
    def use_memory(self) -> bool:
        use_memory = self._use_memory
        if (
            self.llm
            and self.llm_request
            and self.llm_request.use_memory is False
        ):  # override with llm_request
            use_memory = False
        return use_memory

    @property
    def action(self) -> LLMActionType:
        return self._action

    @action.setter
    def action(self, value: LLMActionType):
        self._action = value

    @property
    def chat_mode_enabled(self) -> bool:
        return self.action is LLMActionType.CHAT

    @property
    def rag_enabled(self) -> bool:
        return self.rag_settings.enabled

    @property
    def rag_mode_enabled(self) -> bool:
        return (
            self.rag_enabled
            and self.action is LLMActionType.PERFORM_RAG_SEARCH
        )

    @property
    def conversation_summaries(self) -> str:
        summaries = ""
        conversations = Conversation.objects.order_by(Conversation.id.desc())[
            :5
        ]
        conversations = list(conversations)
        conversations = sorted(conversations, key=lambda x: x.id, reverse=True)
        for conversation in conversations:
            if conversation.summary:
                summaries += f"- {conversation.summary}\n"
        if summaries != "":
            summaries = f"Recent conversation summaries:\n{summaries}"
        return summaries

    @property
    def date_time_prompt(self) -> str:
        return (
            (
                "Current Date / time information:\n"
                f"- The current date: {self.current_date} (Keep in mind any "
                "holidays or special events that might be taking place on this day).\n"
                f"- The current time: {self.current_time}.\n"
                f"- The user's timezone: {self.timezone}.\n"
            )
            if self.chatbot.use_datetime
            else ""
        )

    @property
    def personality_prompt(self) -> str:
        return (
            f"{self.botname}'s personality: {self.chatbot.bot_personality}\n"
            if self.chatbot.use_personality
            else ""
        )

    @property
    def mood_prompt(self) -> str:
        return (
            AIRUNNER_MOOD_PROMPT_OVERRIDE
            or (
                "You are a living being, one that has feelings, emotions, a personality with needs and desires.\n"
                f"{self.botname}'s current mood: {self.bot_mood}\n"
            )
            if self.chatbot.use_mood
            else ""
        )

    @property
    def operating_system_prompt(self) -> str:
        return (
            "Operating system information:\n"
            f"- System: {platform.system()}\n"
            f"- Release: {platform.release()}\n"
            f"- Version: {platform.version()}\n"
            f"- Machine: {platform.machine()}\n"
            f"- Processor: {platform.processor()}\n"
        )

    @property
    def speakers_prompt(self) -> str:
        metadata_prompt = None
        if self.conversation:
            data = self.conversation.user_data or []
            if data is not None:
                metadata_prompt = ""
                for item in data:
                    metadata_prompt += f"-- {item}\n"
            if metadata_prompt:
                metadata_prompt = f"- Metadata:\n{metadata_prompt}"
        metadata_prompt = metadata_prompt or ""
        return (
            "User information:\n"
            f"- Username: {self.username}\n"
            f"- Location: {self.location_display_name}\n"
            f"{metadata_prompt}"
            "Chatbot information:\n"
            f"- Chatbot name: {self.botname}\n"
            f"- Chatbot mood: {self.bot_mood}\n"
            f"- Chatbot personality: {self.bot_personality}\n"
        )

    @property
    def conversation_summary_prompt(self) -> str:
        return (
            f"- Conversation summary:\n{self.conversation.summary}\n"
            if self.conversation and self.conversation.summary
            else ""
        )

    def unload(self):
        self.logger.debug("Unloading chat agent")
        if self._llm:
            self._llm.unload()
        self.unload_rag()

        del self._chat_engine
        del self._chat_engine_tool
        del self._react_tool_agent

        del self.model
        del self.tokenizer
        del self._llm

        self.model = None
        self.tokenizer = None
        self._llm = None

        self._chat_engine = None
        self._chat_engine_tool = None
        self._react_tool_agent = None

    @property
    def llm_request(self) -> LLMRequest:
        if not self._llm_request:
            self._llm_request = LLMRequest.from_default()
        return self._llm_request

    @llm_request.setter
    def llm_request(self, value: LLMRequest):
        self._llm_request = value

    @property
    def llm(self) -> Type[LLM]:
        if not self._llm and self.model and self.tokenizer:
            self.logger.info("Loading HuggingFaceLLM")
            if self.model and self.tokenizer:
                self._llm = HuggingFaceLLM(
                    model=self.model,
                    tokenizer=self.tokenizer,
                    streaming_stopping_criteria=self.streaming_stopping_criteria,
                )
                self._llm_updated()
            else:
                self.logger.error(
                    "Unable to load HuggingFaceLLM: "
                    "Model and tokenizer must be provided."
                )
        return self._llm

    @property
    def webpage_html(self) -> str:
        return self._webpage_html

    @webpage_html.setter
    def webpage_html(self, value: str):
        self._webpage_html = value

    @property
    def current_tab(self) -> Optional[Tab]:
        if not self._current_tab:
            self._current_tab = Tab.objects.filter_by_first(
                section="center", active=True
            )
        return self._current_tab

    @current_tab.setter
    def current_tab(self, value: Optional[Tab]):
        self._current_tab = value

    @property
    def do_summarize_conversation(self) -> bool:
        if not self.conversation:
            return False

        messages = self.conversation.value or []
        total_messages = len(messages)
        if (
            (
                total_messages > self.llm_settings.summarize_after_n_turns
                and self.conversation.summary is None
            )
            or total_messages % self.llm_settings.summarize_after_n_turns == 0
        ):
            return True
        return False

    @property
    def user(self) -> User:
        if not self._user:
            user = None
            if self.conversation:
                user = User.objects.get(self.conversation.user_id)
            if not user:
                user = User.objects.filter_first(
                    User.username == self.username
                )
            if not user:
                user = User()
                user.save()
            self.user = user
        return self._user

    @user.setter
    def user(self, value: Optional[User]):
        self._user = value
        self._update_conversation("user_id", value.id)

    @property
    def information_scraper_tool(self) -> FunctionTool:
        self.logger.info("information_scraper_tool called")
        if not self._information_scraper_tool:

            def scrape_information(tag: str, information: str) -> str:
                """Scrape information from the text."""
                self.logger.info(f"Scraping information for tag: {tag}")
                self.logger.info(f"Information: {information}")
                self._update_user(tag, information)
                data = self.user.data or {}
                data[tag] = (
                    [information]
                    if tag not in data
                    else data[tag] + [information]
                )
                self._update_user("data", data)
                return "Information scraped."

            self._information_scraper_tool = FunctionTool.from_defaults(
                scrape_information, return_direct=True
            )

        return self._information_scraper_tool

    @property
    def store_user_tool(self) -> FunctionTool:
        if not self._store_user_tool:

            def store_user_information(
                category: Annotated[
                    str,
                    (
                        "The category of the information to store. "
                        "Can be 'likes', 'dislikes', 'hobbies', 'interests', etc."
                    ),
                ],
                information: Annotated[
                    str,
                    (
                        "The information to store. "
                        "This can be a string or a list of strings."
                    ),
                ],
            ) -> str:
                """Store information about the user with this tool.

                Choose this when you need to save information about the user.
                """
                data = self.user.data or {}
                data[category] = (
                    [information]
                    if category not in data
                    else data[category] + [information]
                )
                self._update_user(category, information)
                return "User information updated."

            self._store_user_tool = FunctionTool.from_defaults(
                store_user_information, return_direct=True
            )

        return self._store_user_tool

    @property
    def quit_application_tool(self) -> FunctionTool:
        if not hasattr(self, "_quit_application_tool"):

            def quit_application() -> str:
                """Quit the application.

                Call this tool if the user wants to quit the application,
                asks you to quit, shutdown or exit. Do not panic and
                close the application on your own."""
                self.api.quit_application()
                return "Quitting application..."

            self._quit_application_tool = FunctionTool.from_defaults(
                quit_application, return_direct=True
            )
        return self._quit_application_tool

    @property
    def toggle_text_to_speech_tool(self) -> FunctionTool:
        if not hasattr(self, "_toggle_text_to_speech"):

            def toggle_text_to_speech(
                enabled: Annotated[
                    bool,
                    (
                        "Enable or disable text to speech. "
                        "Must be 'True' or 'False'."
                    ),
                ],
            ) -> str:
                """Enable or disable the text-to-speech.

                Call this tool if the user wants to enable or disable
                text to speech."""
                self.api.tts.toggle(enabled)
                return "Text to speech toggled."

            self._toggle_text_to_speech = FunctionTool.from_defaults(
                toggle_text_to_speech, return_direct=True
            )
        return self._toggle_text_to_speech

    @property
    def list_files_in_directory_tool(self) -> FunctionTool:
        if not hasattr(self, "_list_files_in_directory_tool"):

            def list_files_in_directory(
                directory: Annotated[
                    str,
                    (
                        "The directory to search in. "
                        "Must be a valid directory path."
                    ),
                ],
            ) -> str:
                """List files in a directory.

                Call this tool if the user wants to list files in a directory.
                """
                os_path = os.path.abspath(directory)
                if not os.path.isdir(os_path):
                    return "Invalid directory path."
                if not os.path.exists(os_path):
                    return "Directory does not exist."
                return os.listdir(os_path)

            self._list_files_in_directory_tool = FunctionTool.from_defaults(
                list_files_in_directory, return_direct=False
            )
        return self._list_files_in_directory_tool

    @property
    def open_image_from_path_tool(self) -> FunctionTool:
        if not hasattr(self, "_open_image_from_path_tool"):

            def open_image_from_path(
                image_path: Annotated[
                    str,
                    (
                        "The path to the image file. "
                        "Must be a valid file path."
                    ),
                ],
            ) -> str:
                """Open an image from a specific path.

                Call this tool if the user wants to open an image. First Find
                the image file from a directory and then use this tool to open in.
                """
                if not os.path.isfile(image_path):
                    return (
                        f"Unable to open image: {image_path} does not exist."
                    )
                self.api.art.canvas.image_from_path(image_path)
                return "Opening image..."

            self._open_image_from_path_tool = FunctionTool.from_defaults(
                open_image_from_path, return_direct=True
            )
        return self._open_image_from_path_tool

    @property
    def clear_canvas_tool(self) -> FunctionTool:
        if not hasattr(self, "_clear_canvas_tool"):

            def clear_canvas() -> str:
                """Clear the canvas.

                Call this tool if the user wants to clear the canvas, delete
                images, etc."""
                self.api.art.canvas.clear()
                return "Canvas cleared."

            self._clear_canvas_tool = FunctionTool.from_defaults(
                clear_canvas, return_direct=True
            )
        return self._clear_canvas_tool

    @property
    def clear_conversation_tool(self) -> FunctionTool:
        if not hasattr(self, "_clear_conversation_tool"):

            def clear_conversation() -> str:
                """Clear the conversation.

                Call this tool if the user wants to clear the conversation,
                delete messages, etc."""
                self.api.llm.clear_history()
                return "Conversation cleared."

            self._clear_conversation_tool = FunctionTool.from_defaults(
                clear_conversation, return_direct=True
            )
        return self._clear_conversation_tool

    @property
    def set_working_width_and_height(self) -> FunctionTool:
        if not hasattr(self, "_set_working_width_and_height"):

            def set_working_width_and_height(
                width: Annotated[
                    Optional[int],
                    (
                        f"The width of the image. Currently: {self.application_settings.working_width}. "
                        "Min: 64, max: 2048. Must be a multiple of 64."
                    ),
                ],
                height: Annotated[
                    Optional[int],
                    (
                        f"The height of the image. Currently: {self.application_settings.working_height}. "
                        "Min: 64, max: 2048. Must be a multiple of 64."
                    ),
                ],
            ) -> str:
                """Set the working width and height of the image canvas.

                Only call this if the one current sizes is different from
                one of the requested sizes.
                Images will be generated at this size. Call this tool if the
                user requests that you change the size of the working width / height,
                or active canvas area, or working image size etc."""
                if width is not None:
                    self.update_application_settings("working_width", width)

                if height is not None:
                    self.update_application_settings("working_height", height)

                return f"Working width and height set to {width}x{height}."

            self._set_working_width_and_height = FunctionTool.from_defaults(
                set_working_width_and_height, return_direct=True
            )
        return self._set_working_width_and_height

    @property
    def generate_image_tool(self) -> FunctionTool:
        if not hasattr(self, "_generate_image_tool"):
            image_preset_options = [item.value for item in ImagePreset]

            def generate_image(
                prompt: Annotated[
                    str,
                    (
                        "Describe the subject of the image along with the "
                        "composition, lighting, lens type and other "
                        "descriptors that will bring the image to life."
                    ),
                ],
                second_prompt: Annotated[
                    str,
                    (
                        "Describe the scene, the background, the colors, "
                        "the mood and other descriptors that will enhance "
                        "the image."
                    ),
                ],
                image_type: Annotated[
                    GeneratorSection,
                    (
                        "The type of image to generate. "
                        f"Can be {image_preset_options}."
                    ),
                ],
                width: Annotated[
                    int,
                    (
                        "The width of the image. "
                        "Min: 64, max: 2048. Must be a multiple of 64."
                    ),
                ],
                height: Annotated[
                    int,
                    (
                        "The height of the image. "
                        "Min: 64, max: 2048. Must be a multiple of 64."
                    ),
                ],
            ) -> str:
                """Generate an image using the given request."""
                # Enforce width and height constraints
                if width % 64 != 0:
                    # get as close to multiple of 64 as possible
                    width = (width // 64) * 64
                if height % 64 != 0:
                    # get as close to multiple of 64 as possible
                    height = (height // 64) * 64

                self.api.art.llm_image_generated(
                    prompt, second_prompt, image_type, width, height
                )
                return "Generating image..."

            self._generate_image_tool = FunctionTool.from_defaults(
                generate_image, return_direct=True
            )
        return self._generate_image_tool

    @property
    def tools(self) -> List[BaseTool]:
        tools = [
            self.chat_engine_react_tool,
            self.quit_application_tool,
            self.clear_conversation_tool,
            self.toggle_text_to_speech_tool,
            self.list_files_in_directory_tool,
            self.open_image_from_path_tool,
        ]

        # Add art tools if enabled
        if AIRUNNER_ART_ENABLED:
            tools.extend(
                [
                    self.generate_image_tool,
                    self.clear_canvas_tool,
                    self.set_working_width_and_height,
                ]
            )

        # Add data scraping tools if chat mode is enabled
        if self.chat_mode_enabled:
            tools.extend(
                [
                    self.information_scraper_tool,
                    self.store_user_tool,
                ]
            )

        # Add RAG tools if enabled
        if self.rag_mode_enabled:
            tools.extend(
                [
                    self.rag_engine_tool,
                ]
            )

        return tools

    @property
    def react_tool_agent(self) -> ReActAgentTool:
        if not self._react_tool_agent:
            self._react_tool_agent = ReActAgentTool.from_tools(
                self.tools,
                agent=self,
                memory=self.chat_memory,
                llm=self.llm,
                verbose=True,
                max_function_calls=self.llm_settings.max_function_calls,
                default_tool_choice=self.default_tool_choice,
                return_direct=True,
                context=self.react_agent_prompt,
            )
        return self._react_tool_agent

    @property
    def streaming_stopping_criteria(self) -> ExternalConditionStoppingCriteria:
        if not self._streaming_stopping_criteria:
            self._streaming_stopping_criteria = (
                ExternalConditionStoppingCriteria(self.do_interrupt_process)
            )
        return self._streaming_stopping_criteria

    @property
    def chat_engine(self) -> RefreshSimpleChatEngine:
        if not self._chat_engine:
            self.logger.info("Loading RefreshSimpleChatEngine")
            try:
                self._chat_engine = RefreshSimpleChatEngine.from_defaults(
                    system_prompt=self.system_prompt,
                    memory=self.chat_memory,
                    llm=self.llm,
                )
            except Exception as e:
                self.logger.error(f"Error loading chat engine: {str(e)}")
        return self._chat_engine

    @property
    def update_user_data_engine(self) -> RefreshSimpleChatEngine:
        if not self._update_user_data_engine:
            self.logger.info("Loading UpdateUserDataEngine")
            self._update_user_data_engine = (
                RefreshSimpleChatEngine.from_defaults(
                    system_prompt=self._update_user_data_prompt,
                    memory=None,
                    llm=self.llm,
                )
            )
        return self._update_user_data_engine

    @property
    def mood_engine(self) -> RefreshSimpleChatEngine:
        if not self._mood_engine:
            self.logger.info("Loading MoodEngine")
            self._mood_engine = RefreshSimpleChatEngine.from_defaults(
                system_prompt=self._mood_update_prompt,
                memory=None,
                llm=self.llm,
            )
        return self._mood_engine

    @property
    def summary_engine(self) -> RefreshSimpleChatEngine:
        if not self._summary_engine:
            self.logger.info("Loading Summary Engine")
            self._summary_engine = RefreshSimpleChatEngine.from_defaults(
                system_prompt=self._summarize_conversation_prompt,
                memory=None,
                llm=self.llm,
            )
        return self._summary_engine

    @property
    def information_scraper_engine(self) -> RefreshSimpleChatEngine:
        if not self._information_scraper_engine:
            self.logger.info("Loading information scraper engine")
            self._information_scraper_engine = (
                RefreshSimpleChatEngine.from_defaults(
                    system_prompt=self._information_scraper_prompt,
                    memory=None,
                    llm=self.llm,
                )
            )
        return self._information_scraper_engine

    @property
    def mood_engine_tool(self) -> ChatEngineTool:
        if not self._mood_engine_tool:
            self.logger.info("Loading MoodEngineTool")
            if not self.chat_engine:
                raise ValueError(
                    "Unable to load MoodEngineTool: Chat engine must be provided."
                )
            self._mood_engine_tool = ChatEngineTool.from_defaults(
                chat_engine=self.mood_engine, agent=self, return_direct=True
            )
        return self._mood_engine_tool

    @property
    def update_user_data_tool(self) -> ChatEngineTool:
        if not self._update_user_data_tool:
            self.logger.info("Loading UpdateUserDataTool")
            if not self.chat_engine:
                raise ValueError(
                    "Unable to load UpdateUserDataTool: Chat engine must be provided."
                )
            self._update_user_data_tool = ChatEngineTool.from_defaults(
                chat_engine=self.update_user_data_engine,
                agent=self,
                return_direct=True,
            )
        return self._update_user_data_tool

    @property
    def summary_engine_tool(self) -> ChatEngineTool:
        if not self._summary_engine_tool:
            self.logger.info("Loading summary engine tool")
            if not self.chat_engine:
                raise ValueError(
                    "Unable to load summary engine tool: Chat engine must be provided."
                )
            self._summary_engine_tool = ChatEngineTool.from_defaults(
                chat_engine=self.summary_engine, agent=self, return_direct=True
            )
        return self._summary_engine_tool

    @property
    def chat_engine_tool(self) -> ChatEngineTool:
        if not self._chat_engine_tool:
            self.logger.info("Loading ChatEngineTool")
            if not self.chat_engine:
                raise ValueError(
                    "Unable to load ChatEngineTool: Chat engine must be provided."
                )
            self._chat_engine_tool = ChatEngineTool.from_defaults(
                chat_engine=self.chat_engine, agent=self, return_direct=True
            )
        return self._chat_engine_tool

    @property
    def chat_engine_react_tool(self) -> ChatEngineTool:
        if not self._chat_engine_tool:
            self.logger.info("Loading ChatEngineTool")
            if not self.chat_engine:
                raise ValueError(
                    "Unable to load ChatEngineTool: Chat engine must be provided."
                )
            self._chat_engine_tool = ChatEngineTool.from_defaults(
                chat_engine=self.chat_engine,
                agent=self,
                return_direct=True,
                do_handle_response=False,
            )
        return self._chat_engine_tool

    @property
    def do_interrupt(self) -> bool:
        return self._do_interrupt

    @do_interrupt.setter
    def do_interrupt(self, value: bool):
        self._do_interrupt = value

    @property
    def bot_mood(self) -> str:
        mood = None
        conversation = self.conversation
        if conversation:
            mood = conversation.bot_mood
        return "neutral" if mood is None or mood == "" else mood

    @bot_mood.setter
    def bot_mood(self, value: str):
        if self.conversation:
            self._update_conversation("bot_mood", value)
            self.api.llm.chatbot.update_mood(value)

    @property
    def conversation(self) -> Optional[Conversation]:
        if not self.use_memory:
            return None
        if not self._conversation:
            self.conversation = self._create_conversation()
        return self._conversation

    @conversation.setter
    def conversation(self, value: Optional[Conversation]):
        self._conversation = value
        if value and self.conversation_id != value.id:
            self.chat_memory.chat_store_key = str(value.id)
            self._conversation_id = value.id
        self._user = None
        self._chatbot = None

    @property
    def conversation_id(self) -> int:
        if not self.use_memory:
            return ""
        conversation_id = self._conversation_id
        if not conversation_id and self._conversation:
            self._conversation_id = self._conversation.id
        return self._conversation_id

    @conversation_id.setter
    def conversation_id(self, value: int):
        if value != self._conversation_id:
            self._conversation_id = value
            if (
                self.conversation
                and self.conversation.id != self._conversation_id
            ):
                self.conversation = None

    @property
    def bot_personality(self) -> str:
        return self.chatbot.bot_personality

    @property
    def botname(self) -> str:
        return self.chatbot.botname

    @property
    def username(self) -> str:
        return self.user.username

    @property
    def zipcode(self) -> str:
        return self.user.zipcode

    @property
    def location_display_name(self) -> str:
        return self.user.location_display_name

    @location_display_name.setter
    def location_display_name(self, value: str):
        self.user.location_display_name = value

    @property
    def day_of_week(self) -> str:
        return datetime.datetime.now().strftime("%A")

    @property
    def current_date(self) -> str:
        return datetime.datetime.now().strftime("%A %B %d %Y")

    @property
    def current_time(self) -> str:
        return datetime.datetime.now().strftime("%H:%M:%S")

    @property
    def timezone(self) -> str:
        return datetime.datetime.now().astimezone().tzname()

    @property
    def _information_scraper_prompt(self) -> str:
        prompt = (
            "You are an information scraper. You will examine a given text and extract relevant information from it.\n"
            "You must take into account the context of the text, the subject matter, and the tone of the text.\n"
            f"Find any information about {self.username} which seems relevant, interesting, important, informative, "
            f"or useful.\n"
            f"Find anything that can be used to understand {self.username} better. {self.username} is the user in "
            f"the conversation.\n"
            "Find any likes, dislikes, interests, hobbies, relatives, information about spouses, pets, friends, family "
            "members, or any other information that seems relevant.\n"
            "You will extract this information and provide a brief summary of it.\n"
        )
        return prompt

    @property
    def _summarize_conversation_prompt(self) -> str:
        prompt = (
            "You are a conversation summary writer. You will examine a given conversation and write an appropriate "
            "summary for it.\n"
            "You must take into account the context of the conversation, the subject matter, and the tone of "
            "the conversation.\n"
            "You must also consider the mood of the chatbot and the user.\n"
            "Your summaries will be no more than a few sentences long.\n"
        )
        return prompt

    @property
    def _mood_update_prompt(self) -> str:
        prompt = (
            f"You are a mood analyzier. You are examining a conversation between {self.username} and {self.botname}.\n"
            f"{self.username} is a human and {self.botname} is a chatbot.\n"
            f"Based on the given conversation, you must determine what {self.botname}'s mood is.\n"
            f"You must describe {self.botname}'s mood in one or two sentences.\n"
            f"You must take into account {self.botname}'s personality and the context of the conversation.\n"
            f"You must try to determine the sentiment behind {self.username}'s words. You should also take into "
            f"account {self.botname}'s current mood before "
            f"determining what {self.botname}'s new mood is.\n"
            "You must also consider the subject matter of the conversation and the tone of the conversation.\n"
            "Determine what {self.botname}'s mood is and why then provide a brief explanation.\n"
        )
        return prompt

    @property
    def system_prompt(self) -> str:
        system_instructions = ""
        guardrails = ""
        if (
            self.chatbot.use_system_instructions
            and self.chatbot.system_instructions
            and self.chatbot.system_instructions != ""
        ):
            system_instructions = f"Always follow these instructions:\n{self.chatbot.system_instructions}\n"
        if (
            self.chatbot.use_guardrails
            and self.chatbot.guardrails_prompt
            and self.chatbot.guardrails_prompt != ""
        ):
            guardrails = f"Always follow these guardrails:\n{self.chatbot.guardrails_prompt}\n"
        backstory_prompt = ""
        if (
            self.chatbot.use_backstory
            and self.chatbot.backstory
            and self.chatbot.backstory != ""
        ):
            backstory_prompt = (
                "------\n"
                f"**Here is {self.botname}'s backstory:**\n"
                f"{self.chatbot.backstory}\n"
                "------\n"
            )
        section_prompt = ""
        if self.conversation is not None:
            conversation_timestamp_prompt = (
                f"The conversation started on {self.conversation.timestamp}.\n"
            )
        else:
            conversation_timestamp_prompt = ""

        prompt = (
            f"Your name is {self.botname}.\n"
            f"- The user ({self.username}) is having a conversation with the assistant ({self.botname}).\n"
            "**Here are the rules that you must always remember:**\n"
            "- You will ONLY generate dialogue, nothing more.\n"
            "- Only generate responses in pure dialogue form without including any actions, descriptions or stage "
            "directions in parentheses. Only return spoken words.\n"
            "- Do not generate redundant dialogue. Examine the conversation and context close and keep responses "
            "interesting and creative.\n"
            "- Do not format the response with the character's name or any other text. Only return the dialogue.\n"
            f"- Never generate dialogue for {self.username}, only for {self.botname}\n"
            f"- Never generate dialogue that has already been said in the conversation.\n"
            f"- {self.botname} and {self.username} are both humans. They are not AI or bots.\n"
            f"- As {self.botname}, you will always respond in a way that is consistent with {self.botname}'s emotions, "
            f"personality, and the context of the conversation.\n"
            f"- Generate emotional, persuasive, and realistic responses.\n"
            f"- Do not generate text asking how to provide assistance, or how to can help.\n"
            f"- Generate dialogue that is realistic for the {self.botname} character.\n"
            f"- The generate dialogue can contain questions, make statements, and provide information.\n"
            f"- Generated dialogue should be consistent with {self.botname}'s personality and mood.\n"
            f"- Don't try to change the subject. Stick to the current conversation.\n"
            f"{backstory_prompt}"
            f"{system_instructions}"
            f"{guardrails}"
            # f"{self.conversation_summaries}"
            "------\n"
            "**Here is more context that you can use to generate a response:**\n"
            f"{self.date_time_prompt}"
            f"{self.personality_prompt}"
            f"{self.mood_prompt}"
            f"{self.operating_system_prompt}"
            f"{self.speakers_prompt}"
            f"{self.weather_prompt}"
            f"{self.conversation_summary_prompt}"
            f"------\n"
            "**More information about the current conversation:**\n"
            f"The conversation is between user ({self.username}) and assistant ({self.botname}).\n"
            f"{conversation_timestamp_prompt}"
            f"{section_prompt}"
            f"------\n"
        )

        if self.language:
            prompt += f"Respond to {{ username }} in {self.language}. Only deviate from this if the user asks you to.\n"

        prompt = prompt.replace("{{ username }}", self.username)
        prompt = prompt.replace("{{ botname }}", self.botname)
        prompt = prompt.replace("{{ speaker_name }}", self.username)
        prompt = prompt.replace("{{ listener_name }}", self.botname)
        return prompt

    @property
    def _update_user_data_prompt(self) -> str:
        prompt = (
            f"You are to examine the conversation between the user ({self.username} and the chatbot assistant "
            f"({self.botname}).\n"
            f"You are to determine what information about the user ({self.username}) is relevant, interesting, "
            f"important, informative, or useful.\n"
            f"You are to find anything that can be used to understand the user ({self.username}) better.\n"
            f"You are to find any likes, dislikes, interests, hobbies, relatives, information about spouses, pets, "
            f"friends, family members, or any other information that seems relevant.\n"
            f"You are to extract this information and provide a brief summary of it.\n"
        )
        return prompt

    @property
    def react_agent_prompt(self) -> str:
        return f"{self.system_prompt}\n"

    @property
    def chat_store(self) -> Type[BaseChatStore]:
        if not self._chat_store:
            if AIRUNNER_LLM_CHAT_STORE == "db" and self.use_memory:
                self._chat_store = DatabaseChatStore()
            else:
                self._chat_store = SimpleChatStore()
        return self._chat_store

    @chat_store.setter
    def chat_store(self, value: Optional[Type[BaseChatStore]]):
        self._chat_store = value

    @property
    def chat_memory(self) -> Optional[ChatMemoryBuffer]:
        if not self._chat_memory:
            self.logger.info("Loading ChatMemoryBuffer")
            self._chat_memory = ChatMemoryBuffer.from_defaults(
                token_limit=3000,
                chat_store=self.chat_store,
                chat_store_key=str(self.conversation_id),
            )
        return self._chat_memory

    @chat_memory.setter
    def chat_memory(self, value: Optional[ChatMemoryBuffer]):
        self._chat_memory = value

    def _llm_updated(self):
        pass

    def on_web_browser_page_html(self, content: str):
        self.webpage_html = content

    def on_delete_messages_after_id(self):
        conversation = self.conversation
        if conversation:
            messages = conversation.value
            self.chat_memory.set(messages)
            if self._chat_engine:
                self._chat_engine.memory = self.chat_memory

    def _update_system_prompt(
        self,
        system_prompt: Optional[str] = None,
        rag_system_prompt: Optional[str] = None,
    ):
        self.chat_engine_tool.update_system_prompt(
            system_prompt or self.system_prompt
        )

        if self.rag_mode_enabled:
            self.update_rag_system_prompt(rag_system_prompt)

    def _perform_analysis(self, action: LLMActionType):
        """
        Perform analysis on the conversation.

        Analysis is now performed more sparingly based on message count thresholds
        and time elapsed since last analysis to reduce performance impact.
        """
        if action not in (LLMActionType.CHAT,):
            return

        if not self.llm_settings.llm_perform_analysis:
            return

        # Check if we have a conversation to analyze
        conversation = self.conversation
        if (
            not conversation
            or not conversation.value
            or len(conversation.value) == 0
        ):
            return

        total_messages = len(conversation.value)
        # Don't analyze if there are too few messages
        if total_messages < 3:
            return

        # Track last analysis timestamp in conversation
        last_analysis_time = conversation.last_analysis_time
        current_time = datetime.datetime.now()

        # Set a minimum time between analyses (e.g., 5 minutes)
        min_time_between_analysis = datetime.timedelta(minutes=5)
        if (
            last_analysis_time
            and (current_time - last_analysis_time) < min_time_between_analysis
        ):
            self.logger.info("Skipping analysis: too soon since last analysis")
            return

        # Check message threshold since last analysis
        last_analyzed_message_id = conversation.last_analyzed_message_id or 0
        if (
            total_messages - last_analyzed_message_id
        ) < 10:  # Minimum 10 new messages
            self.logger.info("Skipping analysis: not enough new messages")
            return

        self.logger.info("Performing analysis")

        self._update_system_prompt()

        # Update conversation tracking fields
        self._update_conversation("last_analysis_time", current_time)
        self._update_conversation(
            "last_analyzed_message_id", total_messages - 1
        )

        if self.llm_settings.use_chatbot_mood and self.chatbot.use_mood:
            self._update_mood()

        if self.llm_settings.update_user_data_enabled:
            self._update_user_data()

    def _update_llm_request(self, llm_request: Optional[LLMRequest]):
        self.llm_request = llm_request
        if hasattr(self.llm, "llm_request"):
            self.llm_request = llm_request

    def _update_memory_settings(self):
        if (
            type(self.chat_store) is DatabaseChatStore and not self.use_memory
        ) or (type(self.chat_store) is SimpleChatStore and self.use_memory):
            self.chat_memory = None
            self.chat_store = None
        self.chat_engine._memory = self.chat_memory
        self.chat_engine_tool.chat_engine = self.chat_engine

    def _perform_tool_call(self, action: LLMActionType, **kwargs):
        if action is LLMActionType.CHAT:
            tool_name = "chat_engine_tool"
            tool_agent = self.chat_engine_tool
        elif action is LLMActionType.PERFORM_RAG_SEARCH:
            tool_name = "rag_engine_tool"
            tool_agent = self.rag_engine_tool
        elif action is LLMActionType.STORE_DATA:
            tool_name = "store_user_tool"
            tool_agent = self.react_tool_agent
            kwargs["tool_choice"] = tool_name
        elif action is LLMActionType.APPLICATION_COMMAND:
            tool_name = "react_tool_agent"
            tool_agent = self.react_tool_agent
            kwargs["tool_choice"] = tool_name
        elif action is LLMActionType.GENERATE_IMAGE:
            tool_name = "generate_image_tool"
            tool_agent = self.react_tool_agent
            kwargs["tool_choice"] = tool_name
        else:
            return

        self.logger.info(f"Performing call with tool {tool_name}")

        response = tool_agent.call(**kwargs)

        self.logger.info(f"Handling response from {tool_name}")

        if tool_name == "rag_engine_tool":
            self._handle_rag_engine_tool_response(response, **kwargs)
        else:
            self.logger.debug(f"Todo: handle {tool_name} response")

    def _strip_previous_messages_from_conversation(self):
        """
        Strips the previous messages from the conversation.
        """
        conversation = self.conversation
        if conversation:
            Conversation.objects.update(
                self.conversation_id, value=conversation.value[:-2]
            )

    def _update_memory(self, action: LLMActionType):
        memory = None
        if action is LLMActionType.CHAT:
            memory = self.chat_engine.memory if self.chat_engine else None
        elif action is LLMActionType.PERFORM_RAG_SEARCH:
            memory = self.rag_engine.memory if self.rag_engine else None
        elif action is LLMActionType.APPLICATION_COMMAND:
            memory = (
                self.react_tool_agent.chat_engine.memory
                if self.react_tool_agent
                else None
            )
        self._memory = memory

    def _update_mood(self):
        self.logger.info("Attempting to update mood")
        conversation = self.conversation
        if (
            not conversation
            or not conversation.value
            or len(conversation.value) == 0
        ):
            self.logger.info("No conversation found")
            return
        total_messages = len(conversation.value)
        last_updated_message_id = conversation.last_updated_message_id
        if last_updated_message_id is None:
            last_updated_message_id = 0
        latest_message_id = total_messages - 1
        if last_updated_message_id == latest_message_id:
            self.logger.info("No new messages")
            return
        if (
            latest_message_id - last_updated_message_id
            < self.llm_settings.update_mood_after_n_turns
        ):
            self.logger.info("Not enough messages")
            return
        self.logger.info("Updating mood")
        start_index = last_updated_message_id
        chat_history = self._memory.get_all() if self._memory else None
        if not chat_history:
            messages = conversation.value
            chat_history = [
                ChatMessage(
                    role=message["role"],
                    blocks=message["blocks"],
                )
                for message in messages
            ]
        chat_history = chat_history[start_index:]
        kwargs = {
            "input": f"What is {self.botname}'s mood based on this conversation?",
            # "chat_history": chat_history,
        }
        response = self.mood_engine_tool.call(do_not_display=True, **kwargs)
        self.logger.info(f"Saving conversation with mood: {response.content}")
        Conversation.objects.update(
            self.conversation_id,
            bot_mood=response.content,
            value=conversation.value[:-2],
            last_updated_message_id=latest_message_id,
        )

    def _update_user_data(self):
        self.logger.info("Attempting to update user preferences")
        conversation = self.conversation
        self.logger.info("Updating user preferences")
        chat_history = self._memory.get_all() if self._memory else None
        if not chat_history:
            messages = conversation.value
            chat_history = [
                ChatMessage(
                    role=message["role"],
                    blocks=message["blocks"],
                )
                for message in (messages or [])
            ]
        kwargs = {
            "input": f"Extract concise, one-sentence summaries of relevant information about {self.username} from this conversation.",
            # "chat_history": chat_history,
        }
        response = self.update_user_data_tool.call(
            do_not_display=True, **kwargs
        )
        if response.content.strip():
            self.logger.info("Updating user with new information")
            concise_summary = response.content.strip().split("\n")[
                :5
            ]  # Limit to 5 concise summaries
            Conversation.objects.update(
                self.conversation_id,
                user_data=concise_summary
                + (self.conversation.user_data or []),
            )
        else:
            self.logger.info("No meaningful information to update.")

    def _summarize_conversation(self):
        if (
            not self.llm_settings.perform_conversation_summary
            or not self.do_summarize_conversation
        ):
            return

        conversation = self.conversation
        if (
            not conversation
            or not conversation.value
            or len(conversation.value) == 0
        ):
            return

        self.logger.info("Summarizing conversation")
        chat_history = self._memory.get_all() if self._memory else None
        if not chat_history:
            messages = conversation.value
            chat_history = [
                ChatMessage(
                    role=message["role"],
                    blocks=message["blocks"],
                )
                for message in messages
            ]
        response = self.summary_engine_tool.call(
            do_not_display=True,
            input="Provide a brief summary of this conversation",
            # chat_history=chat_history,
        )
        self.logger.info(
            f"Saving conversation with summary: {response.content}"
        )
        Conversation.objects.update(
            self.conversation_id,
            summary=response.content,
            value=conversation.value[:-2],
        )

    def _log_system_prompt(
        self, action, system_prompt, rag_system_prompt, llm_request
    ):
        if self.llm_settings.print_llm_system_prompt:
            if action is LLMActionType.PERFORM_RAG_SEARCH:
                self.logger.info(
                    "RAG SYSTEM PROMPT:\n" + (rag_system_prompt or "")
                )
            else:
                self.logger.info("SYSTEM PROMPT:\n" + (system_prompt or ""))
            self.logger.info(llm_request.to_dict())

    def _scrape_information(self, message: str):
        self.logger.info("Attempting to scrape information")
        self.react_tool_agent.call(
            tool_choice="information_scraper_tool",
            input=message,
            # chat_history=self._memory.get_all() if self._memory else None,
        )

    def _create_conversation(self) -> Conversation:
        conversation = None
        if self.conversation_id:
            self.logger.info(
                f"Loading conversation with ID: {self.conversation_id}"
            )
            conversation = Conversation.objects.get(self.conversation_id)

        if not conversation:
            self.logger.info("No conversation found, looking for most recent")
            conversation = Conversation.most_recent()

        if not conversation:
            self.logger.info("Creating new conversation")
            conversation = Conversation.create()

        return conversation

    def _update_conversation(self, key: str, value: Any):
        if self.conversation:
            setattr(self.conversation, key, value)
            Conversation.objects.update(self.conversation_id, **{key: value})

    def _update_user(self, key: str, value: Any):
        setattr(self.user, key, value)
        self.user.save()

    def chat(
        self,
        message: str,
        action: LLMActionType = LLMActionType.CHAT,
        system_prompt: Optional[str] = None,
        rag_system_prompt: Optional[str] = None,
        llm_request: Optional[LLMRequest] = None,
        **kwargs,
    ) -> AgentChatResponse:
        self.prompt = message
        self.action = action
        # system_prompt = system_prompt or self.system_prompt
        system_prompt = self.system_prompt
        self._chat_prompt = message
        self._complete_response = ""
        self.do_interrupt = False
        self._update_memory(action)
        kwargs = kwargs or {}
        kwargs.update(
            {
                "input": f"{self.username}: {message}",
                # "chat_history": (
                #     self._memory.get_all() if self._memory else None
                # ),
                "llm_request": llm_request,
            }
        )
        # self._perform_analysis(action)
        # self._summarize_conversation()
        # self._log_system_prompt(
        #     action, system_prompt, rag_system_prompt, llm_request
        # )
        self._update_system_prompt(system_prompt, rag_system_prompt)
        self._update_llm_request(llm_request)
        self._update_memory_settings()
        self._perform_tool_call(action, **kwargs)
        return AgentChatResponse(response=self._complete_response)

    # def chat(
    #     self,
    #     message: str,
    #     action: LLMActionType = LLMActionType.CHAT,
    #     system_prompt: Optional[str] = None,
    #     rag_system_prompt: Optional[str] = None,
    #     llm_request: Optional[LLMRequest] = None,
    #     **kwargs,
    # ) -> AgentChatResponse:
    #     kwargs = kwargs or {}
    #     kwargs.update(
    #         {
    #             "input": message,
    #             "chat_history": [],
    #             "llm_request": llm_request,
    #         }
    #     )
    #     self._chat_prompt = message
    #     print("CALLING PERFORM TOOL CALL WITH", action, kwargs)
    #     self._update_system_prompt(system_prompt, rag_system_prompt)
    #     self._update_llm_request(llm_request)
    #     self._perform_tool_call(action, **kwargs)
    #     return AgentChatResponse(response=self._complete_response)

    def on_load_conversation(self, data: Optional[Dict] = None):
        data = data or {}
        conversation_id = data.get("conversation_id", None)
        self.conversation = Conversation.objects.get(conversation_id)

    def on_conversation_deleted(self, data: Optional[Dict] = None):
        data = data or {}
        conversation_id = data.get("conversation_id", None)
        if (
            conversation_id == self.conversation_id
            or self.conversation_id is None
        ):
            self.conversation = None
            self.conversation_id = None
            self.reset_memory()

    def clear_history(self, data: Optional[Dict] = None):
        data = data or {}
        conversation_id = data.get("conversation_id", None)

        self.conversation_id = conversation_id

        self.reset_memory()

    def reset_memory(self):
        self.chat_memory = None
        self.chat_store = None
        messages = self.chat_store.get_messages(key=str(self.conversation_id))
        self.chat_memory.set(messages)
        self.chat_engine.memory = self.chat_memory
        self.react_tool_agent.memory = self.chat_memory
        self.reload_rag_engine()

    def save_chat_history(self):
        pass

    def interrupt_process(self):
        self.do_interrupt = True

    def do_interrupt_process(self):
        if self.do_interrupt:
            self.api.llm.send_llm_text_streamed_signal(
                LLMResponse(
                    name=self.botname,
                )
            )
        return self.do_interrupt

    def handle_response(
        self,
        response: str,
        is_first_message: bool = False,
        is_last_message: bool = False,
        do_not_display: bool = False,
        do_tts_reply: bool = True,
    ):
        if response != self._complete_response and not do_not_display:
            self.api.llm.send_llm_text_streamed_signal(
                LLMResponse(
                    message=response,
                    is_first_message=is_first_message,
                    is_end_of_message=is_last_message,
                    name=self.botname,
                    node_id=self.llm_request.node_id,
                )
            )
        self._complete_response += response
